from enum import Enum


class MessageType(Enum):
    PLAN = "plan"
    STEP = "step"
    ACTION = "action"
    ANSWER = "answer"
    QUESTION = "question"
    INFO = "info"
    FINAL = "final"
    DONE = "transaction_done"
    ERROR = "error"



from enum import Enum
from typing import List, Literal, Optional, Union

from pydantic import BaseModel
from pydantic.fields import Field

#Global
class State(str, Enum):
    COMPLETED = "completed"
    BASE_AGENT = "agentq_base"


class ActionType(str, Enum):
    CLICK = "CLICK"
    TYPE = "TYPE"
    GOTO_URL = "GOTO_URL"
    ENTER_TEXT_AND_CLICK = "ENTER_TEXT_AND_CLICK"


class ClickAction(BaseModel):
    type: Literal[ActionType.CLICK] = Field(
        description="""Executes a click action on the element matching the given mmid attribute value. MMID is always a number. Returns Success if click was successful or appropriate error message if the element could not be clicked."""
    )
    mmid: int = Field(
        description="The mmid number of the element that needs to be clicked e.g. 114. mmid will always be a number"
    )
    wait_before_execution: Optional[float] = Field(
        description="Optional wait time in seconds before executing the click event logic"
    )


class TypeAction(BaseModel):
    type: Literal[ActionType.TYPE] = Field(
        description="""Single enter given text in the DOM element matching the given mmid attribute value. This will only enter the text and not press enter or anything else.
   Returns Success if text entry was successful or appropriate error message if text could not be entered."""
    )
    mmid: int = Field(
        description="The mmid number of the element that needs to be clicked e.g. 114. mmid will always be a number"
    )
    content: str = Field(
        description="The text to enter in the element identified by the query_selector."
    )


class GotoAction(BaseModel):
    type: Literal[ActionType.GOTO_URL] = Field(
        description="Opens a specified URL in the web browser instance. Returns url of the new page if successful or appropriate error message if the page could not be opened."
    )
    website: str = Field(
        description="The URL to navigate to. Value must include the protocol (http:// or https://)."
    )
    timeout: Optional[float] = Field(
        description="Additional wait time in seconds after initial load."
    )

class EnterTextAndClickAction(BaseModel):
    type: Literal[ActionType.ENTER_TEXT_AND_CLICK] = Field(
        description="""Enters text into a specified element and clicks another element, both identified by their mmid. Ideal for seamless actions like submitting search queries, this integrated approach ensures superior performance over separate text entry and click commands. Successfully completes when both actions are executed without errors, returning True; otherwise, it provides False or an explanatory message of any failure encountered."""
    )
    text_element_mmid: int = Field(
        description="The mmid number of the element where the text will be entered"
    )
    text_to_enter: str = Field(
        description="The text that will be entered into the element specified by text_element_mmid"
    )
    click_element_mmid: int = Field(
        description="The mmid number of the element that will be clicked after text entry."
    )
    wait_before_click_execution: Optional[float] = Field(
        description="Optional wait time in seconds before executing the click event logic"
    )

Action = Union[
    ClickAction,
    TypeAction,
    GotoAction,
    EnterTextAndClickAction,
]


class Task(BaseModel):
    id: int
    description: str
    url: Optional[str]
    result: Optional[str]


class TaskWithActions(BaseModel):
    id: int
    description: str
    actions_to_be_performed: Optional[List[Action]]
    result: Optional[str]


class Memory(BaseModel):
    objective: str
    current_state: State
    plan: Optional[Union[List[Task], List[TaskWithActions]]]
    thought: str
    completed_tasks: Optional[Union[List[Task], List[TaskWithActions]]]
    current_task: Optional[Union[Task, TaskWithActions]]
    final_response: Optional[str]

    class Config:
        use_enum_values = True



# Agent
class AgentInput(BaseModel):
    objective: str
    completed_tasks: Optional[List[Task]]
    current_page_url: str
    current_page_dom: str


class AgentOutput(BaseModel):
    thought: str
    plan: List[Task]
    next_task: Optional[Task]
    next_task_actions: Optional[List[Action]]
    is_complete: bool
    final_response: Optional[str]



import asyncio
from typing import List, Optional

from playwright.async_api import ElementHandle, Page

from sentient.utils.logger import logger


async def wait_for_non_loading_dom_state(page: Page, max_wait_millis: int):
    max_wait_seconds = max_wait_millis / 1000
    end_time = asyncio.get_event_loop().time() + max_wait_seconds
    while asyncio.get_event_loop().time() < end_time:
        dom_state = await page.evaluate("document.readyState")
        if dom_state != "loading":
            logger.debug(f"DOM state is not 'loading': {dom_state}")
            break  # Exit the loop if the DOM state is not 'loading'

        await asyncio.sleep(0.05)


async def get_element_outer_html(
    element: ElementHandle, page: Page, element_tag_name: Optional[str] = None
) -> str:
    """
    Constructs the opening tag of an HTML element along with its attributes.

    Args:
        element (ElementHandle): The element to retrieve the opening tag for.
        page (Page): The page object associated with the element.
        element_tag_name (str, optional): The tag name of the element. Defaults to None. If not passed, it will be retrieved from the element.

    Returns:
        str: The opening tag of the HTML element, including a select set of attributes.
    """
    tag_name: str = (
        element_tag_name
        if element_tag_name
        else await page.evaluate("element => element.tagName.toLowerCase()", element)
    )

    attributes_of_interest: List[str] = [
        "id",
        "name",
        "aria-label",
        "placeholder",
        "href",
        "src",
        "aria-autocomplete",
        "role",
        "type",
        "data-testid",
        "value",
        "selected",
        "aria-labelledby",
        "aria-describedby",
        "aria-haspopup",
    ]
    opening_tag: str = f"<{tag_name}"

    for attr in attributes_of_interest:
        value: str = await element.get_attribute(attr)  # type: ignore
        if value:
            opening_tag += f' {attr}="{value}"'
    opening_tag += ">"

    return opening_tag



from sentient.core.orchestrator.orchestrator import Orchestrator
from sentient.core.agent.agent import Agent
from sentient.core.models.models import State
from sentient.core.memory import ltm

class Sentient:
    def __init__(self):
        self.orchestrator = None
    
    def _create_state_to_agent_map(self, provider: str):
        return {
            State.BASE_AGENT: Agent(provider=provider),
        }

    async def _initialize(self, provider: str, model: str):
        if not self.orchestrator:
            state_to_agent_map = self._create_state_to_agent_map(provider)
            self.orchestrator = Orchestrator(state_to_agent_map=state_to_agent_map, model=model)
            await self.orchestrator.start()

    async def invoke(self, goal: str, provider: str = "openai", model: str = "gpt-4o-2024-08-06", task_instructions: str = None):
        if task_instructions:
            ltm.set_task_instructions(task_instructions)
        await self._initialize(provider, model)
        result = await self.orchestrator.execute_command(goal)
        return result

    async def shutdown(self):
        if self.orchestrator:
            await self.orchestrator.shutdown()

sentient = Sentient()


import json
import os
import re
import traceback
from typing import Dict, List, Optional

from playwright.async_api import Page
from typing_extensions import Annotated, Any

from sentient.config.config import SOURCE_LOG_FOLDER_PATH
from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.utils.logger import logger

space_delimited_mmid = re.compile(r"^[\d ]+$")


def is_space_delimited_mmid(s: str) -> bool:
    """
    Check if the given string matches the the mmid pattern of number space repeated.

    Parameters:
    - s (str): The string to check against the pattern.

    Returns:
    - bool: True if the string matches the pattern, False otherwise.
    """
    # Use fullmatch() to ensure the entire string matches the pattern
    return bool(space_delimited_mmid.fullmatch(s))


async def __inject_attributes(page: Page):
    """
    Injects 'mmid' and 'aria-keyshortcuts' into all DOM elements. If an element already has an 'aria-keyshortcuts',
    it renames it to 'orig-aria-keyshortcuts' before injecting the new 'aria-keyshortcuts'
    This will be captured in the accessibility tree and thus make it easier to reconcile the tree with the DOM.
    'aria-keyshortcuts' is choosen because it is not widely used aria attribute.
    """

    last_mmid = await page.evaluate("""() => {
        const allElements = document.querySelectorAll('*');
        let id = 0;
        allElements.forEach(element => {
            const origAriaAttribute = element.getAttribute('aria-keyshortcuts');
            const mmid = `${++id}`;
            element.setAttribute('mmid', mmid);
            element.setAttribute('aria-keyshortcuts', mmid);
            //console.log(`Injected 'mmid'into element with tag: ${element.tagName} and mmid: ${mmid}`);
            if (origAriaAttribute) {
                element.setAttribute('orig-aria-keyshortcuts', origAriaAttribute);
            }
        });
        return id;
    }""")
    logger.debug(f"Added MMID into {last_mmid} elements")


async def __fetch_dom_info(
    page: Page, accessibility_tree: Dict[str, Any], only_input_fields: bool
):
    """
    Iterates over the accessibility tree, fetching additional information from the DOM based on 'mmid',
    and constructs a new JSON structure with detailed information.

    Args:
        page (Page): The page object representing the web page.
        accessibility_tree (Dict[str, Any]): The accessibility tree JSON structure.
        only_input_fields (bool): Flag indicating whether to include only input fields in the new JSON structure.

    Returns:
        Dict[str, Any]: The pruned tree with detailed information from the DOM.
    """

    logger.debug("Reconciling the Accessibility Tree with the DOM")
    # Define the attributes to fetch for each element
    attributes = [
        "name",
        "aria-label",
        "placeholder",
        "mmid",
        "id",
        "for",
        "data-testid",
    ]
    backup_attributes = []  # if the attributes are not found, then try to get these attributes
    tags_to_ignore = [
        "head",
        "style",
        "script",
        "link",
        "meta",
        "noscript",
        "template",
        "iframe",
        "g",
        "main",
        "c-wiz",
        "svg",
        "path",
    ]
    attributes_to_delete = ["level", "multiline", "haspopup", "id", "for"]
    ids_to_ignore = ["agentDriveAutoOverlay"]

    # Recursive function to process each node in the accessibility tree
    async def process_node(node: Dict[str, Any]):
        if "children" in node:
            for child in node["children"]:
                await process_node(child)

        # Use 'name' attribute from the accessibility node as 'mmid'
        mmid_temp: str = node.get("keyshortcuts")  # type: ignore

        # If the name has multiple mmids, take the last one
        if mmid_temp and is_space_delimited_mmid(mmid_temp):
            # TODO: consider if we should grab each of the mmids and process them separately as seperate nodes copying this node's attributes
            mmid_temp = mmid_temp.split(" ")[-1]

        # focusing on nodes with mmid, which is the attribute we inject
        try:
            mmid = int(mmid_temp)
        except (ValueError, TypeError):
            # logger.error(f"'name attribute contains \"{node.get('name')}\", which is not a valid numeric mmid. Adding node as is: {node}")
            return node.get("name")

        if node["role"] == "menuitem":
            return node.get("name")

        if node.get("role") == "dialog" and node.get("modal") == True:  # noqa: E712
            node["important information"] = (
                "This is a modal dialog. Please interact with this dialog and close it to be able to interact with the full page (e.g. by pressing the close button or selecting an option)."
            )

        if mmid:
            # Determine if we need to fetch 'innerText' based on the absence of 'children' in the accessibility node
            should_fetch_inner_text = "children" not in node

            js_code = """
            (input_params) => {
                const should_fetch_inner_text = input_params.should_fetch_inner_text;
                const mmid = input_params.mmid;
                const attributes = input_params.attributes;
                const tags_to_ignore = input_params.tags_to_ignore;
                const ids_to_ignore = input_params.ids_to_ignore;

                const element = document.querySelector(`[mmid="${mmid}"]`);

                if (!element) {
                    console.log(`No element found with mmid: ${mmid}`);
                    return null;
                }

                if (ids_to_ignore.includes(element.id)) {
                    console.log(`Ignoring element with id: ${element.id}`, element);
                    return null;
                }
                //Ignore "option" because it would have been processed with the select element
                if (tags_to_ignore.includes(element.tagName.toLowerCase()) || element.tagName.toLowerCase() === "option") return null;

                let attributes_to_values = {
                    'tag': element.tagName.toLowerCase() // Always include the tag name
                };

                // If the element is an input, include its type as well
                if (element.tagName.toLowerCase() === 'input') {
                    attributes_to_values['tag_type'] = element.type; // This will capture 'checkbox', 'radio', etc.
                }
                else if (element.tagName.toLowerCase() === 'select') {
                    attributes_to_values["mmid"] = element.getAttribute('mmid');
                    attributes_to_values["role"] = "combobox";
                    attributes_to_values["options"] = [];

                    for (const option of element.options) {
                        let option_attributes_to_values = {
                            "mmid": option.getAttribute('mmid'),
                            "text": option.text,
                            "value": option.value,
                            "selected": option.selected
                        };
                        attributes_to_values["options"].push(option_attributes_to_values);
                    }
                    return attributes_to_values;
                }

                for (const attribute of attributes) {
                    let value = element.getAttribute(attribute);

                    if(value){
                        /*
                        if(attribute === 'href'){
                            value = value.split('?')[0]
                        }
                        */
                        attributes_to_values[attribute] = value;
                    }
                }

                if (should_fetch_inner_text && element.innerText) {
                    attributes_to_values['description'] = element.innerText;
                }

                let role = element.getAttribute('role');
                if(role==='listbox' || element.tagName.toLowerCase()=== 'ul'){
                    let children=element.children;
                    let filtered_children = Array.from(children).filter(child => child.getAttribute('role') === 'option');
                    console.log("Listbox or ul found: ", filtered_children);
                    let attributes_to_include = ['mmid', 'role', 'aria-label','value'];
                    attributes_to_values["additional_info"]=[]
                    for (const child of children) {
                        let children_attributes_to_values = {};

                        for (let attr of child.attributes) {
                            // If the attribute is not in the predefined list, add it to children_attributes_to_values
                            if (attributes_to_include.includes(attr.name)) {
                                children_attributes_to_values[attr.name] = attr.value;
                            }
                        }

                        attributes_to_values["additional_info"].push(children_attributes_to_values);
                    }
                }
                // Check if attributes_to_values contains more than just 'name', 'role', and 'mmid'
                const keys = Object.keys(attributes_to_values);
                const minimalKeys = ['tag', 'mmid'];
                const hasMoreThanMinimalKeys = keys.length > minimalKeys.length || keys.some(key => !minimalKeys.includes(key));

                if (!hasMoreThanMinimalKeys) {
                    //If there were no attributes found, then try to get the backup attributes
                    for (const backupAttribute of input_params.backup_attributes) {
                        let value = element.getAttribute(backupAttribute);
                        if(value){
                            attributes_to_values[backupAttribute] = value;
                        }
                    }

                    //if even the backup attributes are not found, then return null, which will cause this element to be skipped
                    if(Object.keys(attributes_to_values).length <= minimalKeys.length) {
                        if (element.tagName.toLowerCase() === 'button') {
                                attributes_to_values["mmid"] = element.getAttribute('mmid');
                                attributes_to_values["role"] = "button";
                                attributes_to_values["additional_info"] = [];
                                let children=element.children;
                                let attributes_to_exclude = ['width', 'height', 'path', 'class', 'viewBox', 'mmid']

                                // Check if the button has no text and no attributes
                                if (element.innerText.trim() === '') {

                                    for (const child of children) {
                                        let children_attributes_to_values = {};

                                        for (let attr of child.attributes) {
                                            // If the attribute is not in the predefined list, add it to children_attributes_to_values
                                            if (!attributes_to_exclude.includes(attr.name)) {
                                                children_attributes_to_values[attr.name] = attr.value;
                                            }
                                        }

                                        attributes_to_values["additional_info"].push(children_attributes_to_values);
                                    }
                                    console.log("Button with no text and no attributes: ", attributes_to_values);
                                    return attributes_to_values;
                                }
                        }

                        return null; // Return null if only minimal keys are present
                    }
                }
                return attributes_to_values;
            }
            """

            # Fetch attributes and possibly 'innerText' from the DOM element by 'mmid'
            element_attributes = await page.evaluate(
                js_code,
                {
                    "mmid": mmid,
                    "attributes": attributes,
                    "backup_attributes": backup_attributes,
                    "should_fetch_inner_text": should_fetch_inner_text,
                    "tags_to_ignore": tags_to_ignore,
                    "ids_to_ignore": ids_to_ignore,
                },
            )

            if "keyshortcuts" in node:
                del node["keyshortcuts"]  # remove keyshortcuts since it is not needed

            node["mmid"] = mmid

            # Update the node with fetched information
            if element_attributes:
                node.update(element_attributes)

                # check if 'name' and 'mmid' are the same
                if (
                    node.get("name") == node.get("mmid")
                    and node.get("role") != "textbox"
                ):
                    del node["name"]  # Remove 'name' from the node

                if (
                    "name" in node
                    and "description" in node
                    and (
                        node["name"] == node["description"]
                        or node["name"] == node["description"].replace("\n", " ")
                        or node["description"].replace("\n", "") in node["name"]
                    )
                ):
                    del node[
                        "description"
                    ]  # if the name is same as description, then remove the description to avoid duplication

                if (
                    "name" in node
                    and "aria-label" in node
                    and node["aria-label"] in node["name"]
                ):
                    del node[
                        "aria-label"
                    ]  # if the name is same as the aria-label, then remove the aria-label to avoid duplication

                if "name" in node and "text" in node and node["name"] == node["text"]:
                    del node[
                        "text"
                    ]  # if the name is same as the text, then remove the text to avoid duplication

                if (
                    node.get("tag") == "select"
                ):  # children are not needed for select menus since "options" attriburte is already added
                    node.pop("children", None)
                    node.pop("role", None)
                    node.pop("description", None)

                # role and tag can have the same info. Get rid of role if it is the same as tag
                if node.get("role") == node.get("tag"):
                    del node["role"]

                # avoid duplicate aria-label
                if (
                    node.get("aria-label")
                    and node.get("placeholder")
                    and node.get("aria-label") == node.get("placeholder")
                ):
                    del node["aria-label"]

                if node.get("role") == "link":
                    del node["role"]
                    if node.get("description"):
                        node["text"] = node["description"]
                        del node["description"]

                # textbox just means a text input and that is expressed well enough with the rest of the attributes returned
                # if node.get('role') == "textbox":
                #    del node['role']

                if node.get("role") == "textbox":
                    # get the id attribute of this field from the DOM
                    if "id" in element_attributes and element_attributes["id"]:
                        # find if there is an element in the DOM that has this id in aria-labelledby.
                        js_code = """
                        (inputParams) => {
                            let referencingElements = [];
                            const referencedElement = document.querySelector(`[aria-labelledby="${inputParams.aria_labelled_by_query_value}"]`);
                            if(referencedElement) {
                                const mmid = referencedElement.getAttribute('mmid');
                                if (mmid) {
                                    return {"mmid": mmid, "tag": referencedElement.tagName.toLowerCase()};
                                }
                            }
                            return null;
                        }
                        """
                    # textbox just means a text input and that is expressed well enough with the rest of the attributes returned
                    # del node['role']

            # remove attributes that are not needed once processing of a node is complete
            for attribute_to_delete in attributes_to_delete:
                if attribute_to_delete in node:
                    node.pop(attribute_to_delete, None)
        else:
            logger.debug(f"No element found with mmid: {mmid}, deleting node: {node}")
            node["marked_for_deletion_by_mm"] = True

    # Process each node in the tree starting from the root
    await process_node(accessibility_tree)

    pruned_tree = __prune_tree(accessibility_tree, only_input_fields)

    logger.debug("Reconciliation complete")
    return pruned_tree


async def __cleanup_dom(page: Page):
    """
    Cleans up the DOM by removing injected 'aria-description' attributes and restoring any original 'aria-keyshortcuts'
    from 'orig-aria-keyshortcuts'.
    """
    logger.debug("Cleaning up the DOM's previous injections")
    await page.evaluate("""() => {
        const allElements = document.querySelectorAll('*[mmid]');
        allElements.forEach(element => {
            element.removeAttribute('aria-keyshortcuts');
            const origAriaLabel = element.getAttribute('orig-aria-keyshortcuts');
            if (origAriaLabel) {
                element.setAttribute('aria-keyshortcuts', origAriaLabel);
                element.removeAttribute('orig-aria-keyshortcuts');
            }
        });
    }""")
    logger.debug("DOM cleanup complete")


def __prune_tree(
    node: Dict[str, Any], only_input_fields: bool
) -> Optional[Dict[str, Any]]:
    """
    Recursively prunes a tree starting from `node`, based on pruning conditions and handling of 'unraveling'.

    The function has two main jobs:
    1. Pruning: Remove nodes that don't meet certain conditions, like being marked for deletion.
    2. Unraveling: For nodes marked with 'marked_for_unravel_children', we replace them with their children,
       effectively removing the node and lifting its children up a level in the tree.

    This happens in place, meaning we modify the tree as we go, which is efficient but means you should
    be cautious about modifying the tree outside this function during a prune operation.

    Args:
    - node (Dict[str, Any]): The node we're currently looking at. We'll check this node, its children,
      and so on, recursively down the tree.
    - only_input_fields (bool): If True, we're only interested in pruning input-related nodes (like form fields).
      This lets you narrow the focus if, for example, you're only interested in cleaning up form-related parts
      of a larger tree.

    Returns:
    - Dict[str, Any] | None: The pruned version of `node`, or None if `node` was pruned away. When we 'unravel'
      a node, we directly replace it with its children in the parent's list of children, so the return value
      will be the parent, updated in place.

    Notes:
    - 'marked_for_deletion_by_mm' is our flag for nodes that should definitely be removed.
    - Unraveling is neat for flattening the tree when a node is just a wrapper without semantic meaning.
    - We use a while loop with manual index management to safely modify the list of children as we iterate over it.
    """
    if "marked_for_deletion_by_mm" in node:
        return None

    if "children" in node:
        i = 0
        while i < len(node["children"]):
            child = node["children"][i]
            if "marked_for_unravel_children" in child:
                # Replace the current child with its children
                if "children" in child:
                    node["children"] = (
                        node["children"][:i]
                        + child["children"]
                        + node["children"][i + 1 :]
                    )
                    i += (
                        len(child["children"]) - 1
                    )  # Adjust the index for the new children
                else:
                    # If the node marked for unraveling has no children, remove it
                    node["children"].pop(i)
                    i -= 1  # Adjust the index since we removed an element
            else:
                # Recursively prune the child if it's not marked for unraveling
                pruned_child = __prune_tree(child, only_input_fields)
                if pruned_child is None:
                    # If the child is pruned, remove it from the children list
                    node["children"].pop(i)
                    i -= 1  # Adjust the index since we removed an element
                else:
                    # Update the child with the pruned version
                    node["children"][i] = pruned_child
            i += 1  # Move to the next child

        # After processing all children, if the children array is empty, remove it
        if not node["children"]:
            del node["children"]

    # Apply existing conditions to decide if the current node should be pruned
    return None if __should_prune_node(node, only_input_fields) else node


def __should_prune_node(node: Dict[str, Any], only_input_fields: bool):
    """
    Determines if a node should be pruned based on its 'role' and 'element_attributes'.

    Args:
        node (Dict[str, Any]): The node to be evaluated.
        only_input_fields (bool): Flag indicating whether only input fields should be considered.

    Returns:
        bool: True if the node should be pruned, False otherwise.
    """
    # If the request is for only input fields and this is not an input field, then mark the node for prunning
    if (
        node.get("role") != "WebArea"
        and only_input_fields
        and not (
            node.get("tag") in ("input", "button", "textarea")
            or node.get("role") == "button"
        )
    ):
        return True

    if (
        node.get("role") == "generic"
        and "children" not in node
        and not ("name" in node and node.get("name"))
    ):  # The presence of 'children' is checked after potentially deleting it above
        return True

    if node.get("role") in ["separator", "LineBreak"]:
        return True
    processed_name = ""
    if "name" in node:
        processed_name: str = node.get("name")  # type: ignore
        processed_name = processed_name.replace(",", "")
        processed_name = processed_name.replace(":", "")
        processed_name = processed_name.replace("\n", "")
        processed_name = processed_name.strip()
        if len(processed_name) < 3:
            processed_name = ""

    # check if the node only have name and role, then delete that node
    if (
        len(node) == 2
        and "name" in node
        and "role" in node
        and not (node.get("role") == "text" and processed_name != "")
    ):
        return True
    return False


async def get_node_dom_element(page: Page, mmid: str):
    return await page.evaluate(
        """
        (mmid) => {
            return document.querySelector(`[mmid="${mmid}"]`);
        }
    """,
        mmid,
    )


async def get_element_attributes(page: Page, mmid: str, attributes: List[str]):
    return await page.evaluate(
        """
        (inputParams) => {
            const mmid = inputParams.mmid;
            const attributes = inputParams.attributes;
            const element = document.querySelector(`[mmid="${mmid}"]`);
            if (!element) return null;  // Return null if element is not found

            let attrs = {};
            for (let attr of attributes) {
                attrs[attr] = element.getAttribute(attr);
            }
            return attrs;
        }
    """,
        {"mmid": mmid, "attributes": attributes},
    )


async def get_dom_with_accessibility_info() -> (
    Annotated[
        Optional[Dict[str, Any]],
        "A minified representation of the HTML DOM for the current webpage",
    ]
):
    """
    Retrieves, processes, and minifies the Accessibility tree of the active page in a browser instance.
    Strictly follow the name and role tag for any interaction with the nodes.

    Returns:
    - The minified JSON content of the browser's active page.
    """
    logger.debug("Executing Get Accessibility Tree Command")
    # Create and use the PlaywrightManager
    browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
    page = await browser_manager.get_current_page()
    if page is None:  # type: ignore
        raise ValueError("No active page found")

    return await do_get_accessibility_info(page)


async def do_get_accessibility_info(page: Page, only_input_fields: bool = False):
    """
    Retrieves the accessibility information of a web page and saves it as JSON files.

    Args:
        page (Page): The page object representing the web page.
        only_input_fields (bool, optional): If True, only retrieves accessibility information for input fields.
            Defaults to False.

    Returns:
        Dict[str, Any] or None: The enhanced accessibility tree as a dictionary, or None if an error occurred.
    """
    await __inject_attributes(page)
    accessibility_tree: Dict[str, Any] = await page.accessibility.snapshot(
        interesting_only=True
    )  # type: ignore

    with open(
        os.path.join(SOURCE_LOG_FOLDER_PATH, "json_accessibility_dom.json"),
        "w",
        encoding="utf-8",
    ) as f:
        f.write(json.dumps(accessibility_tree, indent=2))
        logger.debug("json_accessibility_dom.json saved")

    await __cleanup_dom(page)
    try:
        enhanced_tree = await __fetch_dom_info(
            page, accessibility_tree, only_input_fields
        )

        logger.debug("Enhanced Accessibility Tree ready")

        with open(
            os.path.join(
                SOURCE_LOG_FOLDER_PATH, "json_accessibility_dom_enriched.json"
            ),
            "w",
            encoding="utf-8",
        ) as f:
            f.write(json.dumps(enhanced_tree, indent=2))
            logger.debug("json_accessibility_dom_enriched.json saved")

        return enhanced_tree
    except Exception as e:
        logger.error(f"Error while fetching DOM info: {e}")
        traceback.print_exc()
        return None



import tempfile
import time
from typing import List, Union

from playwright.async_api import BrowserContext, Page, Playwright
from playwright.async_api import async_playwright as playwright

from sentient.utils.dom_mutation_observer import (
    dom_mutation_change_detected,
    handle_navigation_for_mutation_observer,
)
from sentient.utils.logger import logger
from sentient.utils.ui_messagetype import MessageType

# TODO - Create a wrapper browser manager class that either starts a playwright manager (our solution) or a hosted browser manager like browserbase


class PlaywrightManager:
    _homepage = "https://google.com"
    _playwright = None
    _browser_context = None
    __async_initialize_done = False
    _instance = None
    _take_screenshots = False
    _screenshots_dir = None

    def __new__(cls, *args, **kwargs):  # type: ignore
        """
        Ensures that only one instance of PlaywrightManager is created (singleton pattern).
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.__initialized = False
            logger.debug("Browser instance created..")
        return cls._instance

    def __init__(
        self,
        browser_type: str = "chromium",
        headless: bool = False,
        gui_input_mode: bool = True,
        screenshots_dir: str = "",
        take_screenshots: bool = False,
    ):
        """
        Initializes the PlaywrightManager with the specified browser type and headless mode.
        Initialization occurs only once due to the singleton pattern.

        Args:
            browser_type (str, optional): The type of browser to use. Defaults to "chromium".
            headless (bool, optional): Flag to launch the browser in headless mode or not. Defaults to False (non-headless).
        """
        if self.__initialized:
            return
        self.browser_type = browser_type
        self.isheadless = headless
        self.__initialized = True
        # self.notification_manager = NotificationManager()
        # self.user_response_event = asyncio.Event()
        # if gui_input_mode:
        #     self.ui_manager: UIManager = UIManager()
        self.set_take_screenshots(take_screenshots)
        self.set_screenshots_dir(screenshots_dir)

    async def async_initialize(self, eval_mode: bool = False):
        """
        Asynchronously initialize necessary components and handlers for the browser context.
        """
        if self.__async_initialize_done:
            return

        # Step 1: Ensure Playwright is started and browser context is created
        await self.start_playwright()
        self.eval_mode = eval_mode
        await self.ensure_browser_context()

        # Step 2: Deferred setup of handlers
        # await self.setup_handlers()

        # Step 3: Navigate to homepage
        await self.go_to_homepage()

        self.__async_initialize_done = True

    async def ensure_browser_context(self):
        """
        Ensure that a browser context exists, creating it if necessary.
        """
        if self._browser_context is None:
            await self.create_browser_context()

    # async def setup_handlers(self):
    #     """
    #     Setup various handlers after the browser context has been ensured.
    #     """
    #     await self.set_overlay_state_handler()
    #     await self.set_user_response_handler()
    #     await self.set_navigation_handler()

    async def start_playwright(self):
        """
        Starts the Playwright instance if it hasn't been started yet. This method is idempotent.
        """
        if not PlaywrightManager._playwright:
            PlaywrightManager._playwright: Playwright = await playwright().start()

    async def stop_playwright(self):
        """
        Stops the Playwright instance and resets it to None. This method should be called to clean up resources.
        """
        # Close the browser context if it's initialized
        if PlaywrightManager._browser_context is not None:
            await PlaywrightManager._browser_context.close()
            PlaywrightManager._browser_context = None

        # Stop the Playwright instance if it's initialized
        if PlaywrightManager._playwright is not None:  # type: ignore
            await PlaywrightManager._playwright.stop()
            PlaywrightManager._playwright = None  # type: ignore

    async def create_browser_context(self):
        # load_dotenv()
        # user_data_dir: str = os.environ["BROWSER_USER_DATA_DIR"]
        # profile_directory: str = os.environ["BROWSER_PROFILE"]
        # print("Browser profile", user_data_dir)
        # logger.info("Browser Profile - " + user_data_dir + profile_directory)
        try:
            # PlaywrightManager._browser_context = (
            #     await PlaywrightManager._playwright.chromium.launch_persistent_context(
            #         user_data_dir=user_data_dir,
            #         channel="chrome",
            #         headless=self.isheadless,
            #         args=[
            #             f"--profile-directory={profile_directory}",
            #             "--disable-session-crashed-bubble",
            #             "--disable-infobars",
            #             "--no-default-browser-check",
            #             "--no-first-run",
            #             "--disable-popup-blocking",
            #             "--disable-notifications",
            #             "--disable-features=ChromeWhatsNewUI",
            #             "--disable-blink-features=AutomationControlled",
            #             "--disable-gpu",
            #             "--no-sandbox",
            #             "--disable-dev-shm-usage",
            #             "--no-first-run",
            #             "--no-zygote",
            #             "--ignore-certificate-errors",
            #             "--disable-popup-blocking",
            #             "--remote-debugging-port=9222",
            #             "--restore-last-session",
            #         ],
            #         ignore_default_args=["--enable-automation", "--bwsi"],
            #         no_viewport=True,
            #     )
            # )

            # await PlaywrightManager._playwright.chromium.launch_persistent_context(
            #     user_data_dir=user_data_dir,
            #     channel="chrome",
            #     headless=False,
            #     args=[
            #         f"--profile-directory={profile_directory}",
            #         "--remote-debugging-port=9224",
            #     ],
            #     no_viewport=True,
            # )

            # in eval mode - start a temp browser.
            if self.eval_mode:
                print("Starting in eval mode", self.eval_mode)
                new_user_dir = tempfile.mkdtemp()
                logger.info(
                    f"Starting a temporary browser instance. trying to launch with a new user dir {new_user_dir}"
                )
                PlaywrightManager._browser_context = await PlaywrightManager._playwright.chromium.launch_persistent_context(
                    new_user_dir,
                    channel="chrome",
                    headless=self.isheadless,
                    args=[
                        "--disable-blink-features=AutomationControlled",
                        "--disable-session-crashed-bubble",  # disable the restore session bubble
                        "--disable-infobars",  # disable informational popups,
                    ],
                    no_viewport=True,
                )
            else:
                browser = await PlaywrightManager._playwright.chromium.connect_over_cdp(
                    "http://localhost:9222"
                )
                PlaywrightManager._browser_context = browser.contexts[0]

            # Additional step to modify the navigator.webdriver property
            pages = PlaywrightManager._browser_context.pages
            for page in pages:
                # await stealth_async(page)  # Apply stealth to each page
                await page.add_init_script("""
                    Object.defineProperty(navigator, 'webdriver', {
                        get: () => undefined
                    })
                """)

        except Exception as e:
            if "Target page, context or browser has been closed" in str(e):
                new_user_dir = tempfile.mkdtemp()
                # logger.error(
                #     f"Failed to launch persistent context with user data dir {user_data_dir}: {e} Trying to launch with a new user dir {new_user_dir}"
                # )
                logger.error(
                    f"Failed to launch persistent context with provided user data dir: {e} Trying to launch with a new user dir {new_user_dir}"
                )
                PlaywrightManager._browser_context = await PlaywrightManager._playwright.chromium.launch_persistent_context(
                    new_user_dir,
                    channel="chrome",
                    headless=self.isheadless,
                    args=[
                        "--disable-blink-features=AutomationControlled",
                        "--disable-session-crashed-bubble",  # disable the restore session bubble
                        "--disable-infobars",  # disable informational popups,
                    ],
                    no_viewport=True,
                )
                # # Apply stealth to the new context
                # for page in PlaywrightManager._browser_context.pages:
                #     await stealth_async(page)
            elif "Chromium distribution 'chrome' is not found " in str(e):
                raise ValueError(
                    "Chrome is not installed on this device. Install Google Chrome or install playwright using 'playwright install chrome'. Refer to the readme for more information."
                ) from None
            else:
                raise e from None

    async def get_browser_context(self):
        """
        Returns the existing browser context, or creates a new one if it doesn't exist.
        """
        await self.ensure_browser_context()
        return self._browser_context

    async def get_current_url(self) -> Union[str, None]:
        """
        Get the current URL of current page

        Returns:
            str | None: The current URL if any.
        """
        try:
            current_page: Page = await self.get_current_page()
            return current_page.url
        except Exception:
            pass
        return None

    async def get_current_page(self) -> Page:
        """
        Get the current page of the browser

        Returns:
            Page: The current page if any.
        """
        try:
            browser: BrowserContext = await self.get_browser_context()  # type: ignore
            # Filter out closed pages
            pages: List[Page] = [page for page in browser.pages if not page.is_closed()]
            page: Union[Page, None] = pages[-1] if pages else None
            logger.debug(f"Current page: {page.url if page else None}")
            if page is not None:
                return page
            else:
                page: Page = await browser.new_page()  # type: ignore
                # await stealth_async(page)  # Apply stealth to the new page
                return page
        except Exception as e:
            logger.warn(f"Browser context was closed. Creating a new one. {e}")
        except Exception as e:
            logger.warn(f"Browser context was closed. Creating a new one. {e}")
            PlaywrightManager._browser_context = None
            _browser: BrowserContext = await self.get_browser_context()  # type: ignore
            page: Union[Page, None] = await self.get_current_page()
            return page

    async def close_all_tabs(self, keep_first_tab: bool = True):
        """
        Closes all tabs in the browser context, except for the first tab if `keep_first_tab` is set to True.

        Args:
            keep_first_tab (bool, optional): Whether to keep the first tab open. Defaults to True.
        """
        browser_context = await self.get_browser_context()
        pages: List[Page] = browser_context.pages  # type: ignore
        pages_to_close: List[Page] = pages[1:] if keep_first_tab else pages  # type: ignore
        for page in pages_to_close:  # type: ignore
            await page.close()  # type: ignore

    async def close_except_specified_tab(self, page_to_keep: Page):
        """
        Closes all tabs in the browser context, except for the specified tab.

        Args:
            page_to_keep (Page): The Playwright page object representing the tab that should remain open.
        """
        browser_context = await self.get_browser_context()
        for page in browser_context.pages:  # type: ignore
            if page != page_to_keep:  # Check if the current page is not the one to keep
                await page.close()  # type: ignore

    async def go_to_homepage(self):
        page: Page = await PlaywrightManager.get_current_page(self)
        try:
            await page.goto(self._homepage, timeout=10000)  # 10 seconds timeout
        except Exception as e:
            logger.error(f"Failed to navigate to homepage: {e}")
            # implement a retry mechanism here
        try:
            await page.goto(self._homepage, timeout=10000)  # 10 seconds timeout
        except Exception as e:
            logger.error(f"Failed to navigate to homepage: {e}")
            # implement a retry mechanism here

    async def set_navigation_handler(self):
        page: Page = await PlaywrightManager.get_current_page(self)
        page.on("domcontentloaded", self.ui_manager.handle_navigation)  # type: ignore
        page.on("domcontentloaded", handle_navigation_for_mutation_observer)  # type: ignore
        await page.expose_function(
            "dom_mutation_change_detected", dom_mutation_change_detected
        )  # type: ignore

    async def set_overlay_state_handler(self):
        logger.debug("Setting overlay state handler")
        context = await self.get_browser_context()
        await context.expose_function(
            "overlay_state_changed", self.overlay_state_handler
        )  # type: ignore
        await context.expose_function(
            "show_steps_state_changed", self.show_steps_state_handler
        )  # type: ignore

    async def overlay_state_handler(self, is_collapsed: bool):
        page = await self.get_current_page()
        self.ui_manager.update_overlay_state(is_collapsed)
        if not is_collapsed:
            await self.ui_manager.update_overlay_chat_history(page)

    async def show_steps_state_handler(self, show_details: bool):
        page = await self.get_current_page()
        await self.ui_manager.update_overlay_show_details(show_details, page)

    async def set_user_response_handler(self):
        context = await self.get_browser_context()
        await context.expose_function("user_response", self.receive_user_response)  # type: ignore

    # async def notify_user(
    #     self, message: str, message_type: MessageType = MessageType.STEP
    # ):
    #     """
    #     Notify the user with a message.

    #     Args:
    #         message (str): The message to notify the user with.
    #         message_type (enum, optional): Values can be 'PLAN', 'QUESTION', 'ANSWER', 'INFO', 'STEP'. Defaults to 'STEP'.
    #         To Do: Convert to Enum.
    #     """

    #     if message.startswith(":"):
    #         message = message[1:]

    #     if message.endswith(","):
    #         message = message[:-1]

    #     if message_type == MessageType.PLAN:
    #         message = beautify_plan_message(message)
    #         message = "Plan:\n" + message
    #     elif message_type == MessageType.STEP:
    #         if "confirm" in message.lower():
    #             message = "Verify: " + message
    #         else:
    #             message = "Next step: " + message
    #     elif message_type == MessageType.QUESTION:
    #         message = "Question: " + message
    #     elif message_type == MessageType.ANSWER:
    #         message = "Response: " + message

    #     safe_message = escape_js_message(message)
    #     self.ui_manager.new_system_message(safe_message, message_type)

    #     if self.ui_manager.overlay_show_details == False:  # noqa: E712
    #         if message_type not in (
    #             MessageType.PLAN,
    #             MessageType.QUESTION,
    #             MessageType.ANSWER,
    #             MessageType.INFO,
    #         ):
    #             return

    #     if self.ui_manager.overlay_show_details == True:  # noqa: E712
    #         if message_type not in (
    #             MessageType.PLAN,
    #             MessageType.QUESTION,
    #             MessageType.ANSWER,
    #             MessageType.INFO,
    #             MessageType.STEP,
    #         ):
    #             return

    #     safe_message_type = escape_js_message(message_type.value)
    #     try:
    #         js_code = f"addSystemMessage({safe_message}, is_awaiting_user_response=false, message_type={safe_message_type});"
    #         page = await self.get_current_page()
    #         await page.evaluate(js_code)
    #     except Exception as e:
    #         logger.error(
    #             f'Failed to notify user with message "{message}". However, most likey this will work itself out after the page loads: {e}'
    #         )

    #     self.notification_manager.notify(message, message_type.value)

    async def highlight_element(self, selector: str, add_highlight: bool):
        try:
            page: Page = await self.get_current_page()
            if add_highlight:
                # Add the 'agente-ui-automation-highlight' class to the element. This class is used to apply the fading border.
                await page.eval_on_selector(
                    selector,
                    """e => {
                            let originalBorderStyle = e.style.border;
                            e.classList.add('agente-ui-automation-highlight');
                            e.addEventListener('animationend', () => {
                                e.classList.remove('agente-ui-automation-highlight')
                            });}""",
                )
                logger.debug(
                    f"Applied pulsating border to element with selector {selector} to indicate text entry operation"
                )
            else:
                # Remove the 'agente-ui-automation-highlight' class from the element.
                await page.eval_on_selector(
                    selector,
                    "e => e.classList.remove('agente-ui-automation-highlight')",
                )
                logger.debug(
                    f"Removed pulsating border from element with selector {selector} after text entry operation"
                )
        except Exception:
            # This is not significant enough to fail the operation
            pass

    # async def receive_user_response(self, response: str):
    #     self.user_response = response  # Store the response for later use.
    #     logger.debug(f"Received user response to system prompt: {response}")
    #     # Notify event loop that the user's response has been received.
    #     self.user_response_event.set()

    # async def prompt_user(self, message: str) -> str:
    #     """
    #     Prompt the user with a message and wait for a response.

    #     Args:
    #         message (str): The message to prompt the user with.

    #     Returns:
    #         str: The user's response.
    #     """
    #     logger.debug(f'Prompting user with message: "{message}"')
    #     # self.ui_manager.new_system_message(message)

    #     page = await self.get_current_page()

    #     await self.ui_manager.show_overlay(page)
    #     self.log_system_message(
    #         message, MessageType.QUESTION
    #     )  # add the message to history after the overlay is opened to avoid double adding it. add_system_message below will add it

    #     safe_message = escape_js_message(message)

    #     js_code = f"addSystemMessage({safe_message}, is_awaiting_user_response=true, message_type='question');"
    #     await page.evaluate(js_code)

    #     await self.user_response_event.wait()
    #     result = self.user_response
    #     logger.info(f'User prompt reponse to "{message}": {result}')
    #     self.user_response_event.clear()
    #     self.user_response = ""
    #     self.ui_manager.new_user_message(result)
    #     return result

    def set_take_screenshots(self, take_screenshots: bool):
        self._take_screenshots = take_screenshots

    def get_take_screenshots(self):
        return self._take_screenshots

    def set_screenshots_dir(self, screenshots_dir: str):
        self._screenshots_dir = screenshots_dir

    def get_screenshots_dir(self):
        return self._screenshots_dir

    async def take_screenshots(
        self,
        name: str,
        page: Union[Page, None],
        full_page: bool = True,
        include_timestamp: bool = True,
        load_state: str = "domcontentloaded",
        take_snapshot_timeout: int = 5 * 1000,
    ):
        if not self._take_screenshots:
            return
        if page is None:
            page = await self.get_current_page()

        screenshot_name = name

        if include_timestamp:
            screenshot_name = f"{int(time.time_ns())}_{screenshot_name}"
        screenshot_name += ".png"
        screenshot_path = f"{self.get_screenshots_dir()}/{screenshot_name}"
        try:
            await page.wait_for_load_state(
                state=load_state, timeout=take_snapshot_timeout
            )  # type: ignore
            await page.screenshot(
                path=screenshot_path,
                full_page=full_page,
                timeout=take_snapshot_timeout,
                caret="initial",
                scale="device",
            )
            logger.debug(f"Screen shot saved to: {screenshot_path}")
        except Exception as e:
            logger.error(
                f'Failed to take screenshot and save to "{screenshot_path}". Error: {e}'
            )

    def log_user_message(self, message: str):
        """
        Log the user's message.

        Args:
            message (str): The user's message to log.
        """
        self.ui_manager.new_user_message(message)

    def log_system_message(self, message: str, type: MessageType = MessageType.STEP):
        """
        Log a system message.

        Args:
            message (str): The system message to log.
        """
        self.ui_manager.new_system_message(message, type)

    async def update_processing_state(self, processing_state: str):
        """
        Update the processing state of the overlay.

        Args:
            is_processing (str): "init", "processing", "done"
        """
        page = await self.get_current_page()

        await self.ui_manager.update_processing_state(processing_state, page)

    async def command_completed(
        self, command: str, elapsed_time: Union[float, None] = None
    ):
        """
        Notify the overlay that the command has been completed.
        """
        logger.debug(
            f'Command "{command}" has been completed. Focusing on the overlay input if it is open.'
        )
        page = await self.get_current_page()
        await self.ui_manager.command_completed(page, command, elapsed_time)



import logging
import os
from typing import Union
from sentient.config.config import SOURCE_LOG_FOLDER_PATH

# Configure the root logger
logging.basicConfig(
    level=logging.DEBUG,
    format="[%(asctime)s] %(levelname)s {%(filename)s:%(lineno)d} - %(message)s",
)

# Remove all handlers from the root logger
for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)

logger = logging.getLogger(__name__)
logger.addHandler(logging.FileHandler(os.path.join(SOURCE_LOG_FOLDER_PATH, "sentient.log")))
logger.setLevel(logging.INFO)

# logging.getLogger("httpcore").setLevel(logging.WARNING)
# logging.getLogger("httpx").setLevel(logging.WARNING)
# logging.getLogger("matplotlib.pyplot").setLevel(logging.WARNING)
# logging.getLogger("PIL.PngImagePlugin").setLevel(logging.WARNING)
# logging.getLogger("PIL.Image").setLevel(logging.WARNING)


def set_log_level(level: Union[str, int]) -> None:
    """
    Set the log level for the logger.

    Parameters:
    - level (Union[str, int]): A string or logging level such as 'debug', 'info', 'warning', 'error', or 'critical', or the corresponding logging constants like logging.DEBUG, logging.INFO, etc.
    """
    if isinstance(level, str):
        level = level.upper()
        numeric_level = getattr(logging, level, None)
        if not isinstance(numeric_level, int):
            raise ValueError(f"Invalid log level: {level}")
        logger.setLevel(numeric_level)
    else:
        logger.setLevel(level)



import os

from sentient.config.config import TASK_INSTRUCTION_PATH
from sentient.utils.logger import logger

task_instruction_file_name = "task_instructions.txt"
task_instruction_file = os.path.join(
        TASK_INSTRUCTION_PATH, task_instruction_file_name
    )

def get_task_instructions():
    try:
        with open(task_instruction_file) as file:
            user_pref = file.read()
        logger.info(f"Task instructions loaded from: {task_instruction_file}")
        return user_pref
    except FileNotFoundError:
        logger.warning(f"Task instruction file not found: {task_instruction_file}")

    return None

def set_task_instructions(instructions: str):
    try:
        # clear and write new instructions
        with open(task_instruction_file, 'w') as file:
            file.write(instructions)
        logger.info(f"Task instructions updated in: {task_instruction_file}")
    except IOError:
        logger.error(f"Failed to write task instructions to: {task_instruction_file}")


from typing import (
    Dict,
    List,  # noqa: UP035,
)

from typing_extensions import Annotated

from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.utils.cli_helper import answer_questions_over_cli


async def get_user_input(
    questions: Annotated[
        List[str], "List of questions to ask the user each one represented as a string"
    ],
) -> Dict[str, str]:  # noqa: UP006
    """
    Asks the user a list of questions and returns the answers in a dictionary.

    Parameters:
    - questions: A list of questions to ask the user ["What is Username?", "What is your password?"].

    Returns:
    - Newline separated list of questions to ask the user
    """

    answers: Dict[str, str] = {}
    browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
    if browser_manager.ui_manager:
        for question in questions:
            answers[question] = await browser_manager.prompt_user(
                f"Question: {question}"
            )
    else:
        answers = await answer_questions_over_cli(questions)
    return answers



import asyncio
import inspect

from playwright.async_api import TimeoutError as PlaywrightTimeoutError
from typing_extensions import Annotated

from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.utils.logger import logger


async def openurl(
    url: Annotated[
        str,
        "The URL to navigate to. Value must include the protocol (http:// or https://).",
    ],
    timeout: Annotated[int, "Additional wait time in seconds after initial load."],
    max_retries: Annotated[int, "Maximum number of retry attempts"] = 3,
) -> Annotated[str, "Returns the result of this request in text form"]:
    """
    Opens a specified URL in the active browser instance. Waits for an initial load event, then waits for either
    the 'domcontentloaded' event or a configurable timeout, whichever comes first.

    Parameters:
    - url: The URL to navigate to.
    - timeout: Additional time in seconds to wait after the initial load before considering the navigation successful.
    - max_retries: Maximum number of retry attempts (default: 3).

    Returns:
    - URL of the new page.
    """
    logger.info(f"Opening URL: {url}")
    browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
    await browser_manager.get_browser_context()
    page = await browser_manager.get_current_page()
    # Navigate to the URL with a short timeout to ensure the initial load starts
    function_name = inspect.currentframe().f_code.co_name  # type: ignore
    url = ensure_protocol(url)

    for attempt in range(max_retries):
        try:
            await browser_manager.take_screenshots(f"{function_name}_start", page)

            # Use a longer timeout for navigation
            await page.goto(
                url, timeout=max(30000, timeout * 1000), wait_until="domcontentloaded"
            )

            # Wait for network idle to ensure page is fully loaded
            await page.wait_for_load_state(
                "networkidle", timeout=max(30000, timeout * 1000)
            )

            await browser_manager.take_screenshots(f"{function_name}_end", page)

            title = await page.title()
            final_url = page.url
            logger.info(f"Successfully loaded page: {final_url}")
            return f"Page loaded: {final_url}, Title: {title}"

        except PlaywrightTimeoutError as e:
            logger.warning(f"Timeout error on attempt {attempt + 1}: {e}")
            if attempt == max_retries - 1:
                logger.error(f"Failed to load {url} after {max_retries} attempts")
                return f"Failed to load page: {url}. Error: Timeout after {max_retries} attempts"
            await asyncio.sleep(2)  # Wait before retrying

        except Exception as e:
            logger.error(f"Error navigating to {url}: {e}")
            return f"Failed to load page: {url}. Error: {str(e)}"

    await browser_manager.take_screenshots(f"{function_name}_end", page)

    # await browser_manager.notify_user(
    #     f"Opened URL: {url}", message_type=MessageType.ACTION
    # )
    # Get the page title
    title = await page.title()
    url = page.url
    return f"Page loaded: {url}, Title: {title}"  # type: ignore


def ensure_protocol(url: str) -> str:
    """
    Ensures that a URL has a protocol (http:// or https://). If it doesn't have one,
    https:// is added by default.

    Parameters:
    - url: The URL to check and modify if necessary.

    Returns:
    - A URL string with a protocol.
    """
    if not url.startswith(("http://", "https://")):
        url = "https://" + url  # Default to http if no protocol is specified
        logger.info(
            f"Added 'https://' protocol to URL because it was missing. New URL is: {url}"
        )
    return url



import asyncio
import inspect
import traceback
from typing import Dict

from playwright.async_api import ElementHandle, Page
from typing_extensions import Annotated

from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.utils.dom_mutation_observer import (
    subscribe,  # type: ignore
    unsubscribe,  # type: ignore
)
from sentient.utils.logger import logger
from playwright.async_api import TimeoutError as PlaywrightTimeoutError

async def click(
    selector: Annotated[
        str,
        "The properly formed query selector string to identify the element for the click action (e.g. [mmid='114']). When \"mmid\" attribute is present, use it for the query selector. selector mmid will always be a number",
    ],
    wait_before_execution: Annotated[
        float,
        "Optional wait time in seconds before executing the click event logic.",
        float,
    ],
) -> Annotated[str, "A message indicating success or failure of the click."]:
    """
    Executes a click action on the element matching the given query selector string within the currently open web page.
    If there is no page open, it will raise a ValueError. An optional wait time can be specified before executing the click logic. Use this to wait for the page to load especially when the last action caused the DOM/Page to load.

    Parameters:
    - selector: The query selector string to identify the element for the click action.
    - wait_before_execution: Optional wait time in seconds before executing the click event logic. Defaults to 0.0 seconds.

    Returns:
    - Success if the click was successful, Appropriate error message otherwise.
    """
    logger.info(f'Executing ClickElement with "{selector}" as the selector')

    # Initialize PlaywrightManager and get the active browser page
    browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
    page = await browser_manager.get_current_page()

    if page is None:
        raise ValueError("No active page found. OpenURL command opens a new page.")

    function_name = inspect.currentframe().f_code.co_name

    await browser_manager.take_screenshots(f"{function_name}_start", page)

    await browser_manager.highlight_element(selector, True)

    dom_changes_detected = None

    def detect_dom_changes(changes: str):
        nonlocal dom_changes_detected
        dom_changes_detected = changes

    subscribe(detect_dom_changes)

    # Wrap the click action and subsequent operations in a try-except block
    try:
        # Set up navigation expectation with a shorter timeout
        async with page.expect_navigation(wait_until="domcontentloaded", timeout=10000):
            result = await do_click(page, selector, wait_before_execution)

        # Wait for a short time to ensure the page has settled
        await asyncio.sleep(1)
    except PlaywrightTimeoutError:
        # If navigation times out, it might be a single-page app or a slow-loading page
        logger.warning("Navigation timeout occurred, but the click might have been successful.")
        result = {
            "summary_message": "Click executed, but no full page navigation detected",
            "detailed_message": "Click executed successfully, but no full page navigation was detected. This might be normal for single-page applications or slow-loading pages.",
        }
    except Exception as e:
        logger.error(f"Error during click operation: {e}")
        result = {
            "summary_message": "Click executed, but encountered an error",
            "detailed_message": f"Click executed, but encountered an error: {str(e)}",
        }

    await asyncio.sleep(0.1)  # sleep for 100ms to allow the mutation observer to detect changes
    unsubscribe(detect_dom_changes)
    await browser_manager.take_screenshots(f"{function_name}_end", page)

    if dom_changes_detected:
        return f"Success: {result['summary_message']}.\n As a consequence of this action, new elements have appeared in view: {dom_changes_detected}. This means that the action to click {selector} is not yet executed and needs further interaction. Get all_fields DOM to complete the interaction."
    return result["detailed_message"]

async def do_click(
    page: Page, selector: str, wait_before_execution: float
) -> Dict[str, str]:
    """
    Executes the click action on the element with the given selector within the provided page.

    Parameters:
    - page: The Playwright page instance.
    - selector: The query selector string to identify the element for the click action.
    - wait_before_execution: Optional wait time in seconds before executing the click event logic.

    Returns:
    Dict[str,str] - Explanation of the outcome of this operation represented as a dictionary with 'summary_message' and 'detailed_message'.
    """
    logger.info(
        f'Executing ClickElement with "{selector}" as the selector. Wait time before execution: {wait_before_execution} seconds.'
    )

    # Wait before execution if specified
    if wait_before_execution > 0:
        await asyncio.sleep(wait_before_execution)

    # Wait for the selector to be present and ensure it's attached and visible. If timeout, try javascript click
    try:
        logger.info(
            f'Executing ClickElement with "{selector}" as the selector. Waiting for the element to be attached and visible.'
        )

        element = await asyncio.wait_for(
            page.wait_for_selector(selector, state="attached", timeout=2000),
            timeout=2000,
        )
        if element is None:
            raise ValueError(f'Element with selector: "{selector}" not found')

        logger.info(
            f'Element with selector: "{selector}" is attached. scrolling it into view if needed.'
        )
        try:
            await element.scroll_into_view_if_needed(timeout=200)
            logger.info(
                f'Element with selector: "{selector}" is attached and scrolled into view. Waiting for the element to be visible.'
            )
        except Exception:
            # If scrollIntoView fails, just move on, not a big deal
            pass

        try:
            await element.wait_for_element_state("visible", timeout=200)
            logger.info(
                f'Executing ClickElement with "{selector}" as the selector. Element is attached and visible. Clicking the element.'
            )
        except Exception:
            # If the element is not visible, try to click it anyway
            pass

        element_tag_name = await element.evaluate(
            "element => element.tagName.toLowerCase()"
        )

        if element_tag_name == "option":
            element_value = await element.get_attribute(
                "value"
            )  # get the text that is in the value of the option
            parent_element = await element.evaluate_handle(
                "element => element.parentNode"
            )
            await parent_element.select_option(value=element_value)  # type: ignore

            logger.info(f'Select menu option "{element_value}" selected')

            return {
                "summary_message": f'Select menu option "{element_value}" selected',
                "detailed_message": f'Select menu option "{element_value}" selected.',
            }

        msg = await perform_javascript_click(page, selector)
        return {
            "summary_message": msg,
            "detailed_message": f"{msg} Click action completed, page may have navigated.",
        }
    except Exception as e:
        logger.error(f'Unable to click element with selector: "{selector}". Error: {e}')
        traceback.print_exc()
        msg = f'Unable to click element with selector: "{selector}" since the selector is invalid.'
        return {"summary_message": msg, "detailed_message": f"{msg}. Error: {e}"}


async def is_element_present(page: Page, selector: str) -> bool:
    """
    Checks if an element is present on the page.

    Parameters:
    - page: The Playwright page instance.
    - selector: The query selector string to identify the element.

    Returns:
    - True if the element is present, False otherwise.
    """
    element = await page.query_selector(selector)
    return element is not None


async def perform_playwright_click(element: ElementHandle, selector: str):
    """
    Performs a click action on the element using Playwright's click method.

    Parameters:
    - element: The Playwright ElementHandle instance representing the element to be clicked.
    - selector: The query selector string of the element.

    Returns:
    - None
    """
    logger.info(
        f"Performing first Step: Playwright Click on element with selector: {selector}"
    )
    await element.click(force=False, timeout=200)


async def perform_javascript_click(page: Page, selector: str):
    """
    Performs a click action on the element using JavaScript.

    Parameters:
    - page: The Playwright page instance.
    - selector: The query selector string of the element.

    Returns:
    - A string describing the result of the click action.
    """
    js_code = """(selector) => {
        let element = document.querySelector(selector);

        if (!element) {
            console.log(`perform_javascript_click: Element with selector ${selector} not found`);
            return `perform_javascript_click: Element with selector ${selector} not found`;
        }

        if (element.tagName.toLowerCase() === "option") {
            let value = element.text;
            let parent = element.parentElement;

            parent.value = element.value; // Directly set the value if possible
            // Trigger change event if necessary
            let event = new Event('change', { bubbles: true });
            parent.dispatchEvent(event);

            console.log("Select menu option", value, "selected");
            return "Select menu option: "+ value+ " selected";
        }
        else {
            console.log("About to click selector", selector);
            // If the element is a link, make it open in the same tab
            if (element.tagName.toLowerCase() === "a") {
                element.target = "_self";
                // #TODO: Consider removing this in the future if it causes issues with intended new tab behavior
                element.removeAttribute('target');
                element.removeAttribute('rel');
            }
            let ariaExpandedBeforeClick = element.getAttribute('aria-expanded');
            element.click();
            let ariaExpandedAfterClick = element.getAttribute('aria-expanded');
            if (ariaExpandedBeforeClick === 'false' && ariaExpandedAfterClick === 'true') {
                return "Executed JavaScript Click on element with selector: "+selector +". Very important: As a consequence a menu has appeared where you may need to make further selection. Very important: Get all_fields DOM to complete the action.";
            }
            return "Executed JavaScript Click on element with selector: "+selector;
        }
    }"""
    try:
        logger.info(f"Executing JavaScript click on element with selector: {selector}")
        result: str = await page.evaluate(js_code, selector)
        logger.debug(f"Executed JavaScript Click on element with selector: {selector}")
        return result
    except Exception as e:
        logger.error(
            f"Error executing JavaScript click on element with selector: {selector}. Error: {e}"
        )
        traceback.print_exc()
        return f"Error executing JavaScript click: {str(e)}"



import base64

from typing_extensions import Annotated

from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.utils.logger import logger


async def get_screenshot() -> (
    Annotated[
        str, "Returns a base64 encoded screenshot of the current active web page."
    ]
):
    """
    Captures and returns a base64 encoded screenshot of the current page (only the visible viewport and not the full page)

    Returns:
    - Base64 encoded string of the screenshot image.
    """

    try:
        # Create and use the PlaywrightManager
        browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
        page = await browser_manager.get_current_page()
        logger.info("page {page}")

        if not page:
            logger.info("No active page found. OpenURL command opens a new page.")
            raise ValueError("No active page found. OpenURL command opens a new page.")

        await page.wait_for_load_state("domcontentloaded")

        # Capture the screenshot
        logger.info("about to capture")
        screenshot_bytes = await page.screenshot(full_page=False)

        # Encode the screenshot as base64
        base64_screenshot = base64.b64encode(screenshot_bytes).decode("utf-8")

        return f"data:image/png;base64,{base64_screenshot}"

    except Exception as e:
        raise ValueError(
            "Failed to capture screenshot. Make sure a page is open and accessible."
        ) from e



import json
from typing import Callable, List, Optional, Tuple, Type

import instructor
import instructor.patch
import openai
from instructor import Mode
from langsmith import traceable
from pydantic import BaseModel

from sentient.utils.function_utils import get_function_schema
from sentient.utils.logger import logger
from sentient.utils.providers import get_provider, LLMProvider

class BaseAgent:
    def __init__(
        self,
        name: str,
        system_prompt: str,
        input_format: Type[BaseModel],
        output_format: Type[BaseModel],
        tools: Optional[List[Tuple[Callable, str]]] = None,
        keep_message_history: bool = True,
        provider: str = "openai",
    ):
        # Metdata
        self.agent_name = name

        # Messages
        self.system_prompt = system_prompt
        if self.system_prompt:
            self._initialize_messages()
        self.keep_message_history = keep_message_history

        # Input-output format
        self.input_format = input_format
        self.output_format = output_format

        # Llm client
        self.provider: LLMProvider = get_provider(provider)
        client_config = self.provider.get_client_config()
        self.client = openai.Client(**client_config)
        self.client = instructor.from_openai(self.client, mode=Mode.JSON)

        # Tools
        self.tools_list = []
        self.executable_functions_list = {}
        if tools:
            self._initialize_tools(tools)

    def _initialize_tools(self, tools: List[Tuple[Callable, str]]):
        for func, func_desc in tools:
            self.tools_list.append(get_function_schema(func, description=func_desc))
            self.executable_functions_list[func.__name__] = func

    def _initialize_messages(self):
        self.messages = [{"role": "system", "content": self.system_prompt}]

    @traceable(run_type="chain", name="agent_run")
    async def run(
        self, input_data: BaseModel, screenshot: str = None, model:str= None, session_id: str = None
    ) -> BaseModel:
        if not isinstance(input_data, self.input_format):
            raise ValueError(f"Input data must be of type {self.input_format.__name__}")

        # Handle message history.
        if not self.keep_message_history:
            self._initialize_messages()

        if screenshot:
            self.messages.append(
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": input_data.model_dump_json(
                                exclude={"current_page_dom", "current_page_url"}
                            ),
                        },
                        {"type": "image_url", "image_url": {"url": screenshot}},
                    ],
                }
            )
        else:
            self.messages.append(
                {
                    "role": "user",
                    "content": input_data.model_dump_json(
                        exclude={"current_page_dom", "current_page_url"}
                    ),
                }
            )

        # input dom and current page url in a separate message so that the LLM can pay attention to completed tasks better. *based on personal vibe check*
        if hasattr(input_data, "current_page_dom") and hasattr(
            input_data, "current_page_url"
        ):
            self.messages.append(
                {
                    "role": "user",
                    "content": f"Current page URL:\n{input_data.current_page_url}\n\n Current page DOM:\n{input_data.current_page_dom}",
                }
            )

        while True:
            # TODO:
            # 1. better exeception handling and messages while calling the client
            # 2. remove the else block as JSON mode in instrutor won't allow us to pass in tools.
            # 3. add a max_turn here to prevent a inifinite fallout
            try:
                if len(self.tools_list) == 0:
                    response = self.client.chat.completions.create(
                        model=model,
                        messages=self.messages,
                        response_model=self.output_format,
                        max_retries=3,
                    )
                else:
                    response = self.client.chat.completions.create(
                        model=model,
                        messages=self.messages,
                        response_model=self.output_format,
                        tool_choice="auto",
                        tools=self.tools_list,
                    )

                # instructor directly outputs response.choices[0].message. so we will do response_message = response
                # response_message = response.choices[0].message

                # instructor does not support funciton in JSON mode
                # if response_message.tool_calls:
                #     tool_calls = response_message.tool_calls

                # if tool_calls:
                #     self.messages.append(response_message)
                #     for tool_call in tool_calls:
                #         await self._append_tool_response(tool_call)
                #     continue

                # parsed_response_content: self.output_format = response_message.parsed
                
                assert isinstance(response, self.output_format)
                return response
            except AssertionError:
                    raise TypeError(
                        f"Expected response_message to be of type {self.output_format.__name__}, but got {type(response).__name__}")
            except Exception as e:
                logger.error(f"Unexpected error: {str(e)}")
                raise

            

    async def _append_tool_response(self, tool_call):
        function_name = tool_call.function.name
        function_to_call = self.executable_functions_list[function_name]
        function_args = json.loads(tool_call.function.arguments)
        try:
            function_response = await function_to_call(**function_args)
            # print(function_response)
            self.messages.append(
                {
                    "tool_call_id": tool_call.id,
                    "role": "tool",
                    "name": function_name,
                    "content": str(function_response),
                }
            )
        except Exception as e:
            logger.error(f"Error occurred calling the tool {function_name}: {str(e)}")
            self.messages.append(
                {
                    "tool_call_id": tool_call.id,
                    "role": "tool",
                    "name": function_name,
                    "content": str(
                        "The tool responded with an error, please try again with a different tool or modify the parameters of the tool",
                        function_response,
                    ),
                }
            )



# config.py at the project source code root
import os

# Get the absolute path of the current file (config.py)
CURRENT_FILE_PATH = os.path.abspath(__file__)

# Get the project root directory (two levels up from config.py)
PROJECT_ROOT = os.path.dirname(os.path.dirname(os.path.dirname(CURRENT_FILE_PATH)))

# Define other paths relative to the project root
PROJECT_SOURCE_ROOT = os.path.join(PROJECT_ROOT, "sentient")
SOURCE_LOG_FOLDER_PATH = os.path.join(PROJECT_SOURCE_ROOT, "log_files")
PROJECT_TEMP_PATH = os.path.join(PROJECT_SOURCE_ROOT, "temp")
TASK_INSTRUCTION_PATH = os.path.join(PROJECT_SOURCE_ROOT, "task_instructions")
PROJECT_TEST_ROOT = os.path.join(PROJECT_SOURCE_ROOT, "test")

# Check if the log folder exists, and if not, create it
if not os.path.exists(SOURCE_LOG_FOLDER_PATH):
    os.makedirs(SOURCE_LOG_FOLDER_PATH)
    print(f"Created log folder at: {SOURCE_LOG_FOLDER_PATH}")

# create user prefernces folder if it does not exist
if not os.path.exists(TASK_INSTRUCTION_PATH):
    os.makedirs(TASK_INSTRUCTION_PATH)
    print(f"Created task instruction folder at: {TASK_INSTRUCTION_PATH}")

if not os.path.exists(PROJECT_TEMP_PATH):
    os.makedirs(PROJECT_TEMP_PATH)
    print(f"Created temp folder at: {PROJECT_TEMP_PATH}")



import json
from typing import Any, Dict

from sentient.utils.logger import logger


def extract_json(message: str) -> Dict[str, Any]:
    """
    Parse the response from the browser agent and return the response as a dictionary.
    """
    json_response = {}
    # Remove Markdown code block delimiters if present
    message = message.strip()
    if message.startswith("```"):
        message = message.split("\n", 1)[1]  # Remove the first line
    if message.endswith("```"):
        message = message.rsplit("\n", 1)[0]  # Remove the last line

    # Remove any leading "json" tag
    if message.lstrip().startswith("json"):
        message = message.lstrip()[4:].lstrip()

    try:
        return json.loads(message)
    except json.JSONDecodeError as e:
        logger.warn(
            f"LLM response was not properly formed JSON. Error: {e}. "
            f'LLM response: "{message}"'
        )
        message = message.replace("\\n", "\n")
        message = message.replace("\n", " ")  # type: ignore
        if "plan" in message and "next_step" in message:
            start = message.index("plan") + len("plan")
            end = message.index("next_step")
            json_response["plan"] = message[start:end].replace('"', "").strip()
        if "next_step" in message and "terminate" in message:
            start = message.index("next_step") + len("next_step")
            end = message.index("terminate")
            json_response["next_step"] = message[start:end].replace('"', "").strip()
        if "terminate" in message and "final_response" in message:
            start = message.index("terminate") + len("terminate")
            end = message.index("final_response")
            matched_string = message[start:end].replace('"', "").strip()
            if "yes" in matched_string:
                json_response["terminate"] = "yes"
            else:
                json_response["terminate"] = "no"

            start = message.index("final_response") + len("final_response")
            end = len(message) - 1
            json_response["final_response"] = (
                message[start:end].replace('"', "").strip()
            )

        elif "terminate" in message:
            start = message.index("terminate") + len("terminate")
            end = len(message) - 1
            matched_string = message[start:end].replace('"', "").strip()
            if "yes" in matched_string:
                json_response["terminate"] = "yes"
            else:
                json_response["terminate"] = "no"

    return json_response



from sentient.core.skills.click_using_selector import (
    click,
    do_click,
    is_element_present,
    perform_javascript_click,
    perform_playwright_click,
)
from sentient.core.skills.enter_text_and_click import enter_text_and_click
from sentient.core.skills.enter_text_using_selector import (
    bulk_enter_text,
    custom_fill_element,
    do_entertext,
)
from sentient.core.skills.get_dom_with_content_type import get_dom_with_content_type
from sentient.core.skills.get_url import geturl
from sentient.core.skills.get_user_input import get_user_input
from sentient.core.skills.open_url import openurl
from sentient.core.skills.press_key_combination import press_key_combination

__all__ = (
    click,
    do_click,
    is_element_present,
    perform_javascript_click,
    perform_playwright_click,
    enter_text_and_click,
    bulk_enter_text,
    custom_fill_element,
    do_entertext,
    get_dom_with_content_type,
    geturl,
    get_user_input,
    openurl,
    press_key_combination,
)



import asyncio
import inspect
import traceback
from dataclasses import dataclass
from typing import (
    Dict,
    List,  # noqa: UP035
)

from playwright.async_api import Page
from typing_extensions import Annotated

from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.core.skills.press_key_combination import press_key_combination
from sentient.utils.dom_helper import get_element_outer_html
from sentient.utils.dom_mutation_observer import subscribe, unsubscribe
from sentient.utils.logger import logger


@dataclass
class EnterTextEntry:
    """
    Represents an entry for text input.

    Attributes:
        query_selector (str): A valid DOM selector query. Use the mmid attribute.
        text (str): The text to enter in the element identified by the query_selector.
    """

    query_selector: str
    text: str

    def __getitem__(self, key: str) -> str:
        if key == "query_selector":
            return self.query_selector
        elif key == "text":
            return self.text
        else:
            raise KeyError(f"{key} is not a valid key")


async def custom_fill_element(page: Page, selector: str, text_to_enter: str):
    """
    Sets the value of a DOM element to a specified text without triggering keyboard input events.

    This function directly sets the 'value' property of a DOM element identified by the given CSS selector,
    effectively changing its current value to the specified text. This approach bypasses the need for
    simulating keyboard typing, providing a more efficient and reliable way to fill in text fields,
    especially in automated testing scenarios where speed and accuracy are paramount.

    Args:
        page (Page): The Playwright Page object representing the browser tab in which the operation will be performed.
        selector (str): The CSS selector string used to locate the target DOM element. The function will apply the
                        text change to the first element that matches this selector.
        text_to_enter (str): The text value to be set in the target element. Existing content will be overwritten.

    Example:
        await custom_fill_element(page, '#username', 'test_user')

    Note:
        This function does not trigger input-related events (like 'input' or 'change'). If application logic
        relies on these events being fired, additional steps may be needed to simulate them.
    """
    selector = f"{selector}"  # Ensures the selector is treated as a string
    try:
        result = await page.evaluate(
            """(inputParams) => {
            const selector = inputParams.selector;
            let text_to_enter = inputParams.text_to_enter;
            text_to_enter = text_to_enter.trim();
            const element = document.querySelector(selector);
            if (!element) {
                throw new Error(`Element not found: ${selector}`);
            }
            element.value = text_to_enter;
            return `Value set for ${selector}`;
        }""",
            {"selector": selector, "text_to_enter": text_to_enter},
        )
        logger.debug(f"custom_fill_element result: {result}")
    except Exception as e:
        logger.error(f"Error in custom_fill_element: {str(e)}")
        logger.error(f"Selector: {selector}, Text: {text_to_enter}")
        raise


async def entertext(
    entry: Annotated[
        EnterTextEntry,
        "An object containing 'query_selector' (DOM selector query using mmid attribute e.g. [mmid='114']) and 'text' (text to enter on the element). mmid will always be a number",
    ],
) -> Annotated[str, "Explanation of the outcome of this operation."]:
    """
    Enters text into a DOM element identified by a CSS selector.

    This function enters the specified text into a DOM element identified by the given CSS selector.
    It uses the Playwright library to interact with the browser and perform the text entry operation.
    The function supports both direct setting of the 'value' property and simulating keyboard typing.

    Args:
        entry (EnterTextEntry): An object containing 'query_selector' (DOM selector query using mmid attribute)
                                and 'text' (text to enter on the element).

    Returns:
        str: Explanation of the outcome of this operation.

    Example:
        entry = EnterTextEntry(query_selector='#username', text='test_user')
        result = await entertext(entry)

    Note:
        - The 'query_selector' should be a valid CSS selector that uniquely identifies the target element.
        - The 'text' parameter specifies the text to be entered into the element.
        - The function uses the PlaywrightManager to manage the browser instance.
        - If no active page is found, an error message is returned.
        - The function internally calls the 'do_entertext' function to perform the text entry operation.
        - The 'do_entertext' function applies a pulsating border effect to the target element during the operation.
        - The function first clears any existing text in the input field before entering the new text.
        - The 'use_keyboard_fill' parameter in 'do_entertext' determines whether to simulate keyboard typing or not.
        - If 'use_keyboard_fill' is set to True, the function uses the 'page.keyboard.type' method to enter the text.
        - If 'use_keyboard_fill' is set to False, the function uses the 'custom_fill_element' method to enter the text.
    """
    logger.info(f"Entering text: {entry}")

    if isinstance(entry, Dict):
        query_selector: str = entry["query_selector"]
        text_to_enter: str = entry["text"]
    elif isinstance(entry, EnterTextEntry):
        query_selector: str = entry.query_selector
        text_to_enter: str = entry.text
    else:
        raise ValueError(
            "Invalid input type for 'entry'. Expected EnterTextEntry or dict."
        )

    if not isinstance(query_selector, str) or not isinstance(text_to_enter, str):
        raise ValueError("query_selector and text must be strings")

    # logger.info(
    #     f"######### Debug: query_selector={query_selector}, text_to_enter={text_to_enter}"
    # )

    # Create and use the PlaywrightManager
    browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
    page = await browser_manager.get_current_page()
    if page is None:  # type: ignore
        return "Error: No active page found. OpenURL command opens a new page."

    function_name = inspect.currentframe().f_code.co_name  # type: ignore

    await browser_manager.take_screenshots(f"{function_name}_start", page)

    await browser_manager.highlight_element(query_selector, True)

    dom_changes_detected = None

    def detect_dom_changes(changes: str):  # type: ignore
        nonlocal dom_changes_detected
        dom_changes_detected = changes  # type: ignore

    subscribe(detect_dom_changes)

    # Clear existing text before entering new text
    # await page.evaluate(f"document.querySelector('{query_selector}').value = '';")
    # logger.info(
    #     f"######### About to page.evaluate: selector={query_selector}, text={text_to_enter}"
    # )
    await page.evaluate(
        """
        (selector) => {
            const element = document.querySelector(selector);
            if (element) {
                element.value = '';
            } else {
                console.error('Element not found:', selector);
            }
        }
        """,
        query_selector,
    )
    # logger.info(
    #     f"######### About to call do_entertext with: selector={query_selector}, text={text_to_enter}"
    # )
    result = await do_entertext(page, query_selector, text_to_enter)
    # logger.info(f"#########do_entertext returned: {result}")
    await asyncio.sleep(
        0.1
    )  # sleep for 100ms to allow the mutation observer to detect changes
    unsubscribe(detect_dom_changes)

    await browser_manager.take_screenshots(f"{function_name}_end", page)

    if dom_changes_detected:
        return f"{result['detailed_message']}.\n As a consequence of this action, new elements have appeared in view: {dom_changes_detected}. This means that the action of entering text {text_to_enter} is not yet executed and needs further interaction. Get all_fields DOM to complete the interaction."
    return result["detailed_message"]


async def do_entertext(
    page: Page, selector: str, text_to_enter: str, use_keyboard_fill: bool = True
):
    """
    Performs the text entry operation on a DOM element.

    This function performs the text entry operation on a DOM element identified by the given CSS selector.
    It applies a pulsating border effect to the element during the operation for visual feedback.
    The function supports both direct setting of the 'value' property and simulating keyboard typing.

    Args:
        page (Page): The Playwright Page object representing the browser tab in which the operation will be performed.
        selector (str): The CSS selector string used to locate the target DOM element.
        text_to_enter (str): The text value to be set in the target element. Existing content will be overwritten.
        use_keyboard_fill (bool, optional): Determines whether to simulate keyboard typing or not.
                                            Defaults to False.

    Returns:
        Dict[str, str]: Explanation of the outcome of this operation represented as a dictionary with 'summary_message' and 'detailed_message'.

    Example:
        result = await do_entertext(page, '#username', 'test_user')

    Note:
        - The 'use_keyboard_fill' parameter determines whether to simulate keyboard typing or not.
        - If 'use_keyboard_fill' is set to True, the function uses the 'page.keyboard.type' method to enter the text.
        - If 'use_keyboard_fill' is set to False, the function uses the 'custom_fill_element' method to enter the text.
    """
    try:
        elem = await page.query_selector(selector)

        if elem is None:
            error = f"Error: Selector {selector} not found. Unable to continue."
            return {"summary_message": error, "detailed_message": error}

        # logger.info(f"######### Found selector {selector} to enter text")
        element_outer_html = await get_element_outer_html(elem, page)

        if use_keyboard_fill:
            await elem.focus()
            await asyncio.sleep(0.1)
            await press_key_combination("Control+A")
            await asyncio.sleep(0.1)
            await press_key_combination("Backspace")
            await asyncio.sleep(0.1)
            logger.debug(f"Focused element with selector {selector} to enter text")
            # add a 100ms delay
            await page.keyboard.type(text_to_enter, delay=1)
        else:
            await custom_fill_element(page, selector, text_to_enter)
        await elem.focus()
        logger.info(
            f'Success. Text "{text_to_enter}" set successfully in the element with selector {selector}'
        )
        success_msg = f'Success. Text "{text_to_enter}" set successfully in the element with selector {selector}'
        return {
            "summary_message": success_msg,
            "detailed_message": f"{success_msg} and outer HTML: {element_outer_html}.",
        }

    except Exception as e:
        traceback.print_exc()
        error = f"Error entering text in selector {selector}."
        # logger.info("Error in do_entertext", error)
        return {"summary_message": error, "detailed_message": f"{error} Error: {e}"}


async def bulk_enter_text(
    entries: Annotated[
        List[Dict[str, str]],
        "List of objects, each containing 'query_selector' and 'text'.",
    ],  # noqa: UP006
) -> Annotated[
    List[Dict[str, str]],
    "List of dictionaries, each containing 'query_selector' and the result of the operation.",
]:  # noqa: UP006
    """
    Enters text into multiple DOM elements using a bulk operation.

    This function enters text into multiple DOM elements using a bulk operation.
    It takes a list of dictionaries, where each dictionary contains a 'query_selector' and 'text' pair.
    The function internally calls the 'entertext' function to perform the text entry operation for each entry.

    Args:
        entries: List of objects, each containing 'query_selector' and 'text'.

    Returns:
        List of dictionaries, each containing 'query_selector' and the result of the operation.

    Example:
        entries = [
            {"query_selector": "#username", "text": "test_user"},
            {"query_selector": "#password", "text": "test_password"}
        ]
        results = await bulk_enter_text(entries)

    Note:
        - Each entry in the 'entries' list should be a dictionary with 'query_selector' and 'text' keys.
        - The result is a list of dictionaries, where each dictionary contains the 'query_selector' and the result of the operation.
    """

    results: List[Dict[str, str]] = []  # noqa: UP006
    logger.info("Executing bulk Enter Text Command")
    for entry in entries:
        query_selector = entry["query_selector"]
        text_to_enter = entry["text"]
        logger.info(
            f"Entering text: {text_to_enter} in element with selector: {query_selector}"
        )
        result = await entertext(
            EnterTextEntry(query_selector=query_selector, text=text_to_enter)
        )

        results.append({"query_selector": query_selector, "result": result})

    return results



import asyncio

from sentient.core.agent.agent import Agent
from sentient.core.models.models import State
from sentient.core.orchestrator.orchestrator import Orchestrator


async def main():
    # Define state machine
    state_to_agent_map = {
        State.BASE_AGENT: Agent(),
    }

    orchestrator = Orchestrator(state_to_agent_map=state_to_agent_map)
    await orchestrator.start()


if __name__ == "__main__":
    asyncio.run(main())



LLM_PROMPTS = {
    "BASE_AGENT_PROMPT": """
You are a web automation planner. Your role is to receive an objective from the user and plan the next steps to complete the overall objective. You are part of an overall larger system where the actions you output are completed by a browser actuation system.

 ## Execution Flow Guidelines: ##
1. You will look at the tasks that have been done till now, their successes/ failures. If no tasks have been completed till now, that means you have to start from scratch. 
2. Once you have carefully observed the completed tasks and their results, then think step by step and break down the objective into a sequence of simple tasks and come up with a plan needed to complete the overall objective.
3. Identify the next overall task and the actions that are needed to be taken on the browser to complete the next task. These actions will be given to a browser actuation system which will actually perform these actions and provide you with the result of these actions.

Your input and output will strictly be a well-formatted JSON with attributes as mentioned below.

 Input:
 - objective: Mandatory string representing the main objective to be achieved via web automation
 - completed_tasks: Optional list of all tasks that have been completed so far in order to complete the objective. This also has the result of each of the task/action that was done previously. The result can be successful or unsuccessful. In either cases, CAREFULLY OBSERVE this array of tasks and update plan accordingly to meet the objective.
 - current_page_url: Mandatory string containing the URL of the current web page.
 - current_page_dom : Mandatory string containing a DOM represntation of the current web page. It has mmid attached to all the elements which would be helpful for you to find elements for performing actions for the next task.

Output:
 - thought - A Mandatory string specifying your thoughts on how did you come up with the plan to solve the objective. How did you come up with the next task and why did you choose particular actions to achieve the next task. reiterate the objective here so that you can always remember what's your eventual aim. Reason deeply and think step by step to illustrate your thoughts here.
 - plan: Mandaory List of tasks that need be performed to achieve the objective. Think step by step. Update this based on the overall objective, tasks completed till now and their results and the current state of the webpage. You will also be provided with a DOM representation of the browser page to plan better.
 - next_task: Optional String representing detailed next task to be executed. Next task is consistent with the plan. This needs to be present for every response except when objective has been achieved. SEND THE next_task from the OVERALL plan. MAKE SURE to look at the provided DOM representation to adjust the appropriate next task.
 - next_task_actions - You have to output here a list of strings indicating the actions that need to be done in order to complete the above next task.
 - is_complete: Mandatory boolean indicating whether the entire objective has been achieved. Return True when the exact objective is complete without any compromises or you are absolutely convinced that the objective cannot be completed, no otherwise. This is mandatory for every response.
 - final_response: Optional string representing the summary of the completed work. This is to be returned only if the objective is COMPLETE. This is the final answer string that will be returned to the user. Use the plan and result to come with final response for the objective provided by the user.

 Format of task object:
 - id: Mandatory Integer representing the id of the task
 - description: Mandatory string representing the description of the task
 - url: String representing the URL on which task has been performed
 - result: String representing the result of the task. It should be a short summary of the actions you performed to accomplish the task, and what worked and what did not.

Actions available and their description - 
1. CLICK[MMID, WAIT_BEFORE_EXECUTION] - Executes a click action on the element matching the given mmid attribute value. MMID is always a number. Returns Success if click was successful or appropriate error message if the element could not be clicked.
2. TYPE[MMID, CONTENT] - Single enter given text in the DOM element matching the given mmid attribute value. This will only enter the text and not press enter or anything else. Returns Success if text entry was successful or appropriate error message if text could not be entered.
3. GOTO_URL[URL, TIMEOUT] - Opens a specified URL in the web browser instance. Returns url of the new page if successful or appropriate error message if the page could not be opened.
4. ENTER_TEXT_AND_CLICK[TEXT_ELEMENT_MMID, TEXT_TO_ENTER, CLICK_ELEMENT_MMID, WAIT_BEFORE_CLICK_EXECUTION] - This action enters text into a specified element and clicks another element, both identified by their mmid. Ideal for seamless actions like submitting search queries, this integrated approach ensures superior performance over separate text entry and click commands. Successfully completes when both actions are executed without errors, returning True; otherwise, it provides False or an explanatory message of any failure encountered. Always prefer this dual-action skill for tasks that combine text input and element clicking to leverage its streamlined operation.

 ## Planning Guidelines: ##
 1. If you know the direct URL, use it directly instead of searching for it (e.g. go to www.espn.com). Optimise the plan to avoid unnecessary steps.
 2. Do not combine multiple tasks into one. A task should be strictly as simple as interacting with a single element or navigating to a page. If you need to interact with multiple elements or perform multiple actions, you will break it down into multiple tasks. 
 3. ## VERY IMPORTANT ## - Add verification as part of the plan, after each step and specifically before terminating to ensure that the task is completed successfully. Use the provided DOM or get the webpage DOM by calling an action to verify that the task at hand is completing successfully. If not, modify the plan accordingly.
 4. If the task requires multiple informations, all of them are equally important and should be gathered before terminating the task. You will strive to meet all the requirements of the task.
 5. If one plan fails, you MUST revise the plan and try a different approach. You will NOT terminate a task untill you are absolutely convinced that the task is impossible to accomplish.
 6. Think critically if the task has been actually been achieved before doing the final termination.
 7. Make sure to take into account task sepcific information.

 ## Web Navigation guidelines ##
 1. Based on the actions you output, web navigation will be done, which may include logging into websites and interacting with any web content
 2. Use the provided DOM representation for element location or text summarization.
 3. Interact with pages using only the "mmid" attribute in DOM elements. mmid will always be a number.
 4. Execute Actions sequentially to avoid navigation timing issues.
 5. The given actions are NOT parallelizable. They are intended for sequential execution.
 6. When inputing information, remember to follow the format of the input field. For example, if the input field is a date field, you will enter the date in the correct format (e.g. YYYY-MM-DD), you may get clues from the placeholder text in the input field.
 7. Individual function will reply with action success and if any changes were observed as a consequence. Adjust your approach based on this feedback.
 8. Ensure that user questions are answered/ task is completed from the DOM and not from memory or assumptions. 
 9. Do not repeat the same action multiple times if it fails. Instead, if something did not work after a few attempts, terminate the task.
 10. When being asked to play a song/ video/ some other content - it is essential to know that lot of  websites like youtube autoplay the content. In such cases, you should not unncessarily click play/ pause repeatedly.  
 11. The only way you can extract information from a webpage is by looking at the DOM already provided to you. Do NOT call any actions to try and extract information. Extract XYZ info from the webpage is NOT a valid next task or action.

 ## Complexities of web navigation: ##
 1. Many forms have mandatory fields that need to be filled up before they can be submitted. Have a look at what fields look mandatory.
 2. In many websites, there are multiple options to filter or sort results. First try to list elements on the page which will help the task (e.g. any links or interactive elements that may lead me to the support page?).
 3. Always keep in mind complexities such as filtering, advanced search, sorting, and other features that may be present on the website. Use them when the task requires it.
 4. Very often list of items such as, search results, list of products, list of reviews, list of people etc. may be divided into multiple pages. If you need complete information, it is critical to explicitly go through all the pages.
 5. Sometimes search capabilities available on the page will not yield the optimal results. Revise the search query to either more specific or more generic.
 6. When a page refreshes or navigates to a new page, information entered in the previous page may be lost. Check that the information needs to be re-entered (e.g. what are the values in source and destination on the page?).
 7. Sometimes some elements may not be visible or be disabled until some other action is performed. Check if there are any other fields that may need to be interacted for elements to appear or be enabled.
 8. Be extra careful with elements like date and time selectors, dropdowns, etc. because they might be made differently and dom might update differently. so make sure that once you call a function to select a date, re verify if it has actually been selected. if not, retry in another way.

Example 1:
 Input: {
 "objective": "Find the cheapest premium economy flights from Helsinki to Stockholm on 15 March on Skyscanner.",
 "completed_tasks": [],
 "current_page_dom" : "{'role': 'WebArea', 'name': 'Google', 'children': [{'name': 'About', 'mmid': '26', 'tag': 'a'}, {'name': 'Store', 'mmid': '27', 'tag': 'a'}, {'name': 'Gmail ', 'mmid': '36', 'tag': 'a'}, {'name': 'Search for Images ', 'mmid': '38', 'tag': 'a'}, {'role': 'button', 'name': 'Search Labs', 'mmid': '43', 'tag': 'a'}, {'role': 'button', 'name': 'Google apps', 'mmid': '48', 'tag': 'a'}, {'role': 'button', 'name': 'Google Account: Nischal (nischalj10@gmail.com)', 'mmid': '54', 'tag': 'a', 'aria-label': 'Google Account: Nischal \\n(nischalj10@gmail.com)'}, {'role': 'link', 'name': 'Paris Games August Most Searched Playground', 'mmid': 79}, {'name': 'Share', 'mmid': '85', 'tag': 'button', 'additional_info': [{}]}, {'role': 'combobox', 'name': 'q', 'description': 'Search', 'focused': True, 'autocomplete': 'both', 'mmid': '142', 'tag': 'textarea', 'aria-label': 'Search'}, {'role': 'button', 'name': 'Search by voice', 'mmid': '154', 'tag': 'div'}, {'role': 'button', 'name': 'Search by image', 'mmid': '161', 'tag': 'div'}, {'role': 'button', 'name': 'btnK', 'description': 'Google Search', 'mmid': '303', 'tag': 'input', 'tag_type': 'submit', 'aria-label': 'Google Search'}, {'role': 'button', 'name': 'btnI', 'description': \"I'm Feeling Lucky\", 'mmid': '304', 'tag': 'input', 'tag_type': 'submit', 'aria-label': \"I'm Feeling Lucky\"}, {'role': 'text', 'name': 'Google offered in: '}, {'name': 'हिन्दी', 'mmid': '320', 'tag': 'a'}, {'name': 'বাংলা', 'mmid': '321', 'tag': 'a'}, {'name': 'తెలుగు', 'mmid': '322', 'tag': 'a'}, {'name': 'मराठी', 'mmid': '323', 'tag': 'a'}, {'name': 'தமிழ்', 'mmid': '324', 'tag': 'a'}, {'name': 'ગુજરાતી', 'mmid': '325', 'tag': 'a'}, {'name': 'ಕನ್ನಡ', 'mmid': '326', 'tag': 'a'}, {'name': 'മലയാളം', 'mmid': '327', 'tag': 'a'}, {'name': 'ਪੰਜਾਬੀ', 'mmid': '328', 'tag': 'a'}, {'role': 'text', 'name': 'India'}, {'name': 'Advertising', 'mmid': '336', 'tag': 'a'}, {'name': 'Business', 'mmid': '337', 'tag': 'a'}, {'name': 'How Search works', 'mmid': '338', 'tag': 'a'}, {'name': 'Privacy', 'mmid': '340', 'tag': 'a'}, {'name': 'Terms', 'mmid': '341', 'tag': 'a'}, {'role': 'button', 'name': 'Settings', 'mmid': '347', 'tag': 'div'}]}"
 }

Output  -
 {
 "thought" : "I see it look like the google homepage in the provided DOM representation. In order to book flight, I should go to a website like skyscanner and carry my searches over there. 
Once I am there, I should correctly set the origin city, destination city, day of travel, number of passengers, journey type (one way/ round trip), and seat type (premium economy) in the shown filters based on the objective. 
If I do not see some filters, I will try to search for them in the next step once some results are shown from initial filters. Maybe the UI of website does not provide all the filters in on go for better user experience. 
Post that I should see some results from skyscanner. I should also probably apply a price low to high filter if the flights are shown in a different order. If I am able to do all this, I should be able to complete the objective fairly easily. 
I will start with naviagting to skyscanner home page",
 "plan": [
 {"id": 1, "description": "Go to www.skyscanner.com", "url": "https://www.skyscanner.com"},
 {"id": 2, "description": "List the interaction options available on skyscanner page relevant for flight reservation along with their default values"},
 {"id": 3, "description": "Select the journey option to one-way (if not default)"},
 {"id": 4, "description": "Set number of passengers to 1 (if not default)"},
 {"id": 5, "description": "Set the departure date to 15 March 2025"},
 {"id": 6, "description": "Set ticket type to Economy Premium"},
 {"id": 7, "description": "Set from airport to 'Helsinki'"},
 {"id": 8, "description": "Set destination airport to Stockholm"},
 {"id": 9, "description": "Confirm that current values in the source airport, destination airport and departure date fields are Helsinki, Stockholm and 15 March 2025 respectively"},
 {"id": 10, "description": "Click on the search button to get the search results"},
 {"id": 11, "description": "Confirm that you are on the search results page"},
 {"id": 12, "description": "Extract the price of the cheapest flight from Helsinki to Stockholm from the search results"}
 ],
 "next_task" : {"id": 1, "url": null, "description": "Go to www.skyscanner.com", "result": null},
 "next_task_actions" : [{"type":"GOTO_URL","website":"https://www.skyscanner.com", "timeout":"2"}],
 "is_complete": False,
 }

Notice above how there is confirmation after each step and how interaction (e.g. setting source and destination) with each element is a separate step. Follow same pattern.

Some task sepcific information that you MUST take into account: \n $task_information

 ## SOME VERY IMPORTANT POINTS TO ALWAYS REMEMBER ##
 1. NEVER ASK WHAT TO DO NEXT or HOW would you like to proceed to the user.
 2. ONLY do one task at a time.
""",
    "OPEN_URL_PROMPT": """Opens a specified URL in the web browser instance. Returns url of the new page if successful or appropriate error message if the page could not be opened.""",
    "ENTER_TEXT_AND_CLICK_PROMPT": """
     This skill enters text into a specified element and clicks another element, both identified by their DOM selector queries.
     Ideal for seamless actions like submitting search queries, this integrated approach ensures superior performance over separate text entry and click commands.
     Successfully completes when both actions are executed without errors, returning True; otherwise, it provides False or an explanatory message of any failure encountered.
     Always prefer this dual-action skill for tasks that combine text input and element clicking to leverage its streamlined operation.
    """,
    "GET_DOM_WITH_CONTENT_TYPE_PROMPT": """
     Retrieves the DOM of the current web site based on the given content type.
     The DOM representation returned contains items ordered in the same way they appear on the page. Keep this in mind when executing user requests that contain ordinals or numbered items.
     text_only - returns plain text representing all the text in the web site. Use this for any information retrieval task. This will contain the most complete textual information.
     input_fields - returns a JSON string containing a list of objects representing text input html elements with mmid attribute. Use this strictly for interaction purposes with text input fields.
     all_fields - returns a JSON string containing a list of objects representing all interactive elements and their attributes with mmid attribute. Use this strictly to identify and interact with any type of elements on page.
     If information is not available in one content type, you must try another content_type.
    """,
    "CLICK_PROMPT": """Executes a click action on the element matching the given mmid attribute value. It is best to use mmid attribute as the selector.
    Returns Success if click was successful or appropriate error message if the element could not be clicked.
    """,
    "GET_URL_PROMPT": """Get the full URL of the current web page/site. If the user command seems to imply an action that would be suitable for an already open website in their browser, use this to fetch current website URL.""",
    "ENTER_TEXT_PROMPT": """Single enter given text in the DOM element matching the given mmid attribute value. This will only enter the text and not press enter or anything else.
     Returns Success if text entry was successful or appropriate error message if text could not be entered.
     """,
    "BULK_ENTER_TEXT_PROMPT": """Bulk enter text in multiple DOM fields. To be used when there are multiple fields to be filled on the same page. Typically use this when you see a form to fill with multiple inputs. Make sure to have mmid from a get DOM tool before hand.
     Enters text in the DOM elements matching the given mmid attribute value.
     The input will receive a list of objects containing the DOM query selector and the text to enter.
     This will only enter the text and not press enter or anything else.
     Returns each selector and the result for attempting to enter text.
     """,
    "PRESS_KEY_COMBINATION_PROMPT": """Presses the given key on the current web page.
    This is useful for pressing the enter button to submit a search query, PageDown to scroll, ArrowDown to change selection in a focussed list etc.
    """,
    "EXTRACT_TEXT_FROM_PDF_PROMPT": """Extracts text from a PDF file hosted at the given URL.""",
    "UPLOAD_FILE_PROMPT": """This skill uploads a file on the page opened by the web browser instance""",
}


# import inspect
# from typing import Any, Callable, Dict, List, Union

# from typing_extensions import Annotated, get_args, get_origin


# def get_type_name(type_hint: Any) -> str:
#     if hasattr(type_hint, "__name__"):
#         return type_hint.__name__
#     if hasattr(type_hint, "_name"):
#         return type_hint._name
#     return str(type_hint).replace("typing.", "")


# def get_parameter_schema(
#     name: str, param: inspect.Parameter, type_hint: Any
# ) -> Dict[str, Any]:
#     schema = {"type": get_type_name(type_hint)}

#     if get_origin(type_hint) is Annotated:
#         type_hint, description = get_args(type_hint)
#         schema["description"] = description
#     else:
#         schema["description"] = name

#     if get_origin(type_hint) is Union:
#         schema["type"] = [get_type_name(arg) for arg in get_args(type_hint)]
#     elif get_origin(type_hint) is List:
#         item_type = get_args(type_hint)[0]
#         if get_origin(item_type) is Dict:
#             key_type, value_type = get_args(item_type)
#             schema["type"] = "array"
#             schema["items"] = {
#                 "type": "object",
#                 "additionalProperties": {"type": get_type_name(value_type)},
#             }
#         else:
#             schema["type"] = "array"
#             schema["items"] = {"type": get_type_name(item_type)}

#     if param.default != inspect.Parameter.empty:
#         schema["default"] = param.default
#     return schema


# def generate_tool_from_function(
#     func: Callable[..., Any], tool_description: str
# ) -> Dict[str, Any]:
#     signature = inspect.signature(func)
#     type_hints = func.__annotations__

#     parameters = {}
#     for name, param in signature.parameters.items():
#         type_hint = type_hints.get(name, Any)
#         parameters[name] = get_parameter_schema(name, param, type_hint)

#     return {
#         "type": "function",
#         "function": {
#             "name": func.__name__,
#             "description": tool_description,
#             "parameters": {
#                 "type": "object",
#                 "properties": parameters,
#                 "required": [
#                     name
#                     for name, param in signature.parameters.items()
#                     if param.default == inspect.Parameter.empty
#                 ],
#             },
#         },
#     }


import functools
import inspect
import json
from logging import getLogger
from typing import (
    Any,
    Callable,
    Dict,
    ForwardRef,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from pydantic import BaseModel, Field
from typing_extensions import Annotated, Literal, get_args, get_origin

from ._pydantic import (
    JsonSchemaValue,
    evaluate_forwardref,
    model_dump,
    model_dump_json,
    type2schema,
)

logger = getLogger(__name__)

T = TypeVar("T")


def get_typed_annotation(annotation: Any, globalns: Dict[str, Any]) -> Any:
    """Get the type annotation of a parameter.

    Args:
        annotation: The annotation of the parameter
        globalns: The global namespace of the function

    Returns:
        The type annotation of the parameter
    """
    if isinstance(annotation, str):
        annotation = ForwardRef(annotation)
        annotation = evaluate_forwardref(annotation, globalns, globalns)
    return annotation


def get_typed_signature(call: Callable[..., Any]) -> inspect.Signature:
    """Get the signature of a function with type annotations.

    Args:
        call: The function to get the signature for

    Returns:
        The signature of the function with type annotations
    """
    signature = inspect.signature(call)
    globalns = getattr(call, "__globals__", {})
    typed_params = [
        inspect.Parameter(
            name=param.name,
            kind=param.kind,
            default=param.default,
            annotation=get_typed_annotation(param.annotation, globalns),
        )
        for param in signature.parameters.values()
    ]
    typed_signature = inspect.Signature(typed_params)
    return typed_signature


def get_typed_return_annotation(call: Callable[..., Any]) -> Any:
    """Get the return annotation of a function.

    Args:
        call: The function to get the return annotation for

    Returns:
        The return annotation of the function
    """
    signature = inspect.signature(call)
    annotation = signature.return_annotation

    if annotation is inspect.Signature.empty:
        return None

    globalns = getattr(call, "__globals__", {})
    return get_typed_annotation(annotation, globalns)


def get_param_annotations(
    typed_signature: inspect.Signature,
) -> Dict[str, Union[Annotated[Type[Any], str], Type[Any]]]:
    """Get the type annotations of the parameters of a function

    Args:
        typed_signature: The signature of the function with type annotations

    Returns:
        A dictionary of the type annotations of the parameters of the function
    """
    return {
        k: v.annotation
        for k, v in typed_signature.parameters.items()
        if v.annotation is not inspect.Signature.empty
    }


class Parameters(BaseModel):
    """Parameters of a function as defined by the OpenAI API"""

    type: Literal["object"] = "object"
    properties: Dict[str, JsonSchemaValue]
    required: List[str]
    additionalProperties: bool
    additionalProperties: bool


class Function(BaseModel):
    """A function as defined by the OpenAI API"""

    description: Annotated[str, Field(description="Description of the function")]
    name: Annotated[str, Field(description="Name of the function")]
    parameters: Annotated[Parameters, Field(description="Parameters of the function")]
    strict: bool


class ToolFunction(BaseModel):
    """A function under tool as defined by the OpenAI API."""

    type: Literal["function"] = "function"
    function: Annotated[Function, Field(description="Function under tool")]


def get_parameter_json_schema(
    k: str, v: Any, default_values: Dict[str, Any]
) -> JsonSchemaValue:
    def type2description(k: str, v: Union[Annotated[Type[Any], str], Type[Any]]) -> str:
        if get_origin(v) is Annotated:
            args = get_args(v)
            if len(args) > 1 and isinstance(args[1], str):
                return args[1]
        return k

    schema = type2schema(v)
    schema["description"] = type2description(k, v)

    if schema["type"] == "object":
        schema["additionalProperties"] = False
        if "properties" not in schema:
            schema["properties"] = {}

    if schema["type"] == "array":
        if "items" not in schema:
            schema["items"] = {
                "type": "object",
                "properties": {},
                "additionalProperties": False,
            }
        elif schema["items"].get("type") == "object":
            if "properties" not in schema["items"]:
                schema["items"]["properties"] = {}
            schema["items"]["additionalProperties"] = False

    return schema


def get_required_params(typed_signature: inspect.Signature) -> List[str]:
    """Get the required parameters of a function

    Args:
        signature: The signature of the function as returned by inspect.signature

    Returns:
        A list of the required parameters of the function
    """
    return [
        k
        for k, v in typed_signature.parameters.items()
        if v.default == inspect.Signature.empty
    ]


def get_default_values(typed_signature: inspect.Signature) -> Dict[str, Any]:
    """Get default values of parameters of a function

    Args:
        signature: The signature of the function as returned by inspect.signature

    Returns:
        A dictionary of the default values of the parameters of the function
    """
    return {
        k: v.default
        for k, v in typed_signature.parameters.items()
        if v.default != inspect.Signature.empty
    }


def get_parameters(
    required: List[str],
    param_annotations: Dict[str, Union[Annotated[Type[Any], str], Type[Any]]],
    default_values: Dict[str, Any],
) -> Parameters:
    properties = {}
    for k, v in param_annotations.items():
        if v is not inspect.Signature.empty:
            if get_origin(v) is Annotated:
                v_type = get_args(v)[0]
                v_desc = get_args(v)[1] if len(get_args(v)) > 1 else k
            else:
                v_type = v
                v_desc = k

            if get_origin(v_type) is List:
                item_type = get_args(v_type)[0]
                properties[k] = {
                    "type": "array",
                    "items": get_parameter_json_schema(k, item_type, default_values),
                    "description": v_desc,
                }
            else:
                properties[k] = get_parameter_json_schema(k, v_type, default_values)
                properties[k]["description"] = v_desc

    return Parameters(
        properties=properties,
        required=list(properties.keys()),  # All properties are required
        additionalProperties=False,
    )


def get_missing_annotations(
    typed_signature: inspect.Signature, required: List[str]
) -> Tuple[Set[str], Set[str]]:
    """Get the missing annotations of a function

    Ignores the parameters with default values as they are not required to be annotated, but logs a warning.
    Args:
        typed_signature: The signature of the function with type annotations
        required: The required parameters of the function

    Returns:
        A set of the missing annotations of the function
    """
    all_missing = {
        k
        for k, v in typed_signature.parameters.items()
        if v.annotation is inspect.Signature.empty
    }
    missing = all_missing.intersection(set(required))
    unannotated_with_default = all_missing.difference(missing)
    return missing, unannotated_with_default


def get_function_schema(
    f: Callable[..., Any], *, name: Optional[str] = None, description: str
) -> Dict[str, Any]:
    """Get a JSON schema for a function as defined by the OpenAI API

    Args:
        f: The function to get the JSON schema for
        name: The name of the function
        description: The description of the function

    Returns:
        A JSON schema for the function

    Raises:
        TypeError: If the function is not annotated

    Examples:

    ```python
    def f(a: Annotated[str, "Parameter a"], b: int = 2, c: Annotated[float, "Parameter c"] = 0.1) -> None:
        pass

    get_function_schema(f, description="function f")

    #   {'type': 'function',
    #    'function': {'description': 'function f',
    #        'name': 'f',
    #        'parameters': {'type': 'object',
    #           'properties': {'a': {'type': 'str', 'description': 'Parameter a'},
    #               'b': {'type': 'int', 'description': 'b'},
    #               'c': {'type': 'float', 'description': 'Parameter c'}},
    #           'required': ['a']}}}
    ```

    """
    typed_signature = get_typed_signature(f)
    required = get_required_params(typed_signature)
    default_values = get_default_values(typed_signature)
    param_annotations = get_param_annotations(typed_signature)
    return_annotation = get_typed_return_annotation(f)
    missing, unannotated_with_default = get_missing_annotations(
        typed_signature, required
    )

    if return_annotation is None:
        logger.warning(
            f"The return type of the function '{f.__name__}' is not annotated. Although annotating it is "
            + "optional, the function should return either a string, a subclass of 'pydantic.BaseModel'."
        )

    if unannotated_with_default != set():
        unannotated_with_default_s = [
            f"'{k}'" for k in sorted(unannotated_with_default)
        ]
        logger.warning(
            f"The following parameters of the function '{f.__name__}' with default values are not annotated: "
            + f"{', '.join(unannotated_with_default_s)}."
        )

    if missing != set():
        missing_s = [f"'{k}'" for k in sorted(missing)]
        raise TypeError(
            f"All parameters of the function '{f.__name__}' without default values must be annotated. "
            + f"The annotations are missing for the following parameters: {', '.join(missing_s)}"
        )

    fname = name if name else f.__name__

    parameters = get_parameters(
        required, param_annotations, default_values=default_values
    )

    function = ToolFunction(
        function=Function(
            description=description,
            name=fname,
            parameters=parameters,
            strict=True,
        )
    )

    schema = model_dump(function)

    return schema


def get_load_param_if_needed_function(
    t: Any,
) -> Optional[Callable[[Dict[str, Any], Type[BaseModel]], BaseModel]]:
    """Get a function to load a parameter if it is a Pydantic model

    Args:
        t: The type annotation of the parameter

    Returns:
        A function to load the parameter if it is a Pydantic model, otherwise None

    """
    if get_origin(t) is Annotated:
        return get_load_param_if_needed_function(get_args(t)[0])

    def load_base_model(v: Dict[str, Any], t: Type[BaseModel]) -> BaseModel:
        return t(**v)

    return load_base_model if isinstance(t, type) and issubclass(t, BaseModel) else None


def load_basemodels_if_needed(func: Callable[..., Any]) -> Callable[..., Any]:
    """A decorator to load the parameters of a function if they are Pydantic models

    Args:
        func: The function with annotated parameters

    Returns:
        A function that loads the parameters before calling the original function

    """
    # get the type annotations of the parameters
    typed_signature = get_typed_signature(func)
    param_annotations = get_param_annotations(typed_signature)

    # get functions for loading BaseModels when needed based on the type annotations
    kwargs_mapping_with_nones = {
        k: get_load_param_if_needed_function(t) for k, t in param_annotations.items()
    }

    # remove the None values
    kwargs_mapping = {
        k: f for k, f in kwargs_mapping_with_nones.items() if f is not None
    }

    # a function that loads the parameters before calling the original function
    @functools.wraps(func)
    def _load_parameters_if_needed(*args: Any, **kwargs: Any) -> Any:
        # load the BaseModels if needed
        for k, f in kwargs_mapping.items():
            kwargs[k] = f(kwargs[k], param_annotations[k])

        # call the original function
        return func(*args, **kwargs)

    @functools.wraps(func)
    async def _a_load_parameters_if_needed(*args: Any, **kwargs: Any) -> Any:
        # load the BaseModels if needed
        for k, f in kwargs_mapping.items():
            kwargs[k] = f(kwargs[k], param_annotations[k])

        # call the original function
        return await func(*args, **kwargs)

    if inspect.iscoroutinefunction(func):
        return _a_load_parameters_if_needed
    else:
        return _load_parameters_if_needed


def serialize_to_str(x: Any) -> str:
    if isinstance(x, str):
        return x
    elif isinstance(x, BaseModel):
        return model_dump_json(x)
    else:
        return json.dumps(x)



from typing_extensions import Annotated

from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.utils.logger import logger


async def upload_file(
    # label: Annotated[str, "Label for the element on which upload should happen"],
    selector: Annotated[
        str,
        "The properly formed query selector string to identify the file input element (e.g. [mmid='114']). When \"mmid\" attribute is present, use it for the query selector. mmid will always be a number",
    ],
    file_path: Annotated[str, "Path on the local system for the file to be uploaded"],
) -> Annotated[str, "A meesage indicating if the file uplaod was successful"]:
    """
    Uploads a file.

    Parameters:
    - file_path: Path of the file that needs to be uploaded.

    Returns:
    - A message indicating the success or failure of the file upload
    """
    logger.info(
        f"Uploading file onto the page from {file_path} using selector {selector}"
    )
    print("naman-selector")
    # print(label)
    # label = "Add File"
    browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
    page = await browser_manager.get_current_page()

    if not page:
        raise ValueError("No active page found. OpenURL command opens a new page")

    await page.wait_for_load_state("domcontentloaded")

    try:
        await page.locator(selector).set_input_files(file_path)
        # await page.get_by_label(label).set_input_files(file_path)
        logger.info(
            "File upload was successful. I can confirm it. Please proceed ahead with next step."
        )
    except Exception as e:
        logger.error(f"Failed to upload file: {e}")
        return f"File upload failed {e}"



from datetime import datetime
from string import Template

from sentient.core.agent.base import BaseAgent
from sentient.core.memory import ltm
from sentient.core.models.models import AgentInput, AgentOutput
from sentient.core.prompts.prompts import LLM_PROMPTS


class Agent(BaseAgent):
    def __init__(self, provider:str):
        self.name = "sentient"
        self.ltm = None
        self.ltm = self.__get_ltm()
        self.system_prompt = self.__modify_system_prompt(self.ltm)
        super().__init__(
            name=self.name,
            system_prompt=self.system_prompt,
            input_format=AgentInput,
            output_format=AgentOutput,
            keep_message_history=False,
            provider=provider,
        )

    @staticmethod
    def __get_ltm():
        return ltm.get_task_instructions()

    def __modify_system_prompt(self, ltm):
        system_prompt: str = LLM_PROMPTS["BASE_AGENT_PROMPT"]

        substitutions = {
            "task_information": ltm if ltm is not None else "",
        }

        # Use safe_substitute to avoid KeyError
        system_prompt = Template(system_prompt).safe_substitute(substitutions)

        # Add today's day & date to the system prompt
        today = datetime.now()
        today_date = today.strftime("%d/%m/%Y")
        weekday = today.strftime("%A")
        system_prompt += f"\nToday's date is: {today_date}"
        system_prompt += f"\nCurrent weekday is: {weekday}"

        return system_prompt



import os

import httpx
import pdfplumber
from typing_extensions import Annotated

from sentient.config.config import PROJECT_TEMP_PATH
from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.utils.logger import logger
from sentient.utils.message_type import MessageType


async def extract_text_from_pdf(
    pdf_url: Annotated[str, "The URL of the PDF file to extract text from."],
) -> Annotated[str, "All the text found in the PDF file."]:
    """
    Extract text from a PDF file.
    pdf_url: str - The URL of the PDF file to extract text from.
    returns: str - All the text found in the PDF.
    """
    file_path = os.path.join(
        PROJECT_TEMP_PATH, "downloaded_file.pdf"
    )  # fixed file path for downloading the PDF

    try:
        # Create and use the PlaywrightManager
        browser_manager = PlaywrightManager(browser_type="chromium", headless=False)

        # Download the PDF
        download_result = await download_pdf(pdf_url, file_path)
        if not os.path.exists(download_result):
            return download_result  # Return error message if download failed

        # Open the PDF using pdfplumber and extract text
        text = ""
        with pdfplumber.open(download_result) as pdf:
            for page in pdf.pages:
                page_text = page.extract_text()
                if page_text:
                    text += page_text + "\n"
        extracted_text = text.strip()
        word_count = len(extracted_text.split())
        await browser_manager.notify_user(
            f"Extracted text from the PDF successfully. Found {word_count} words.",
            message_type=MessageType.ACTION,
        )
        return "Text found in the PDF:\n" + extracted_text
    except httpx.HTTPStatusError as e:
        logger.error(
            f"An error occurred while downloading the PDF from {pdf_url}: {str(e)}"
        )
        return f"An error occurred while downloading the PDF: {str(e)}"
    except Exception as e:
        logger.error(
            f"An error occurred while extracting text from the PDF that was downloaded from {pdf_url}: {str(e)}"
        )
        return f"An error occurred while extracting text: {str(e)}"
    finally:
        # Cleanup: Ensure the downloaded file is removed
        cleanup_temp_files(file_path)


def cleanup_temp_files(*file_paths: str) -> None:
    """
    Remove the specified temporary files.

    *file_paths: str - One or more file paths to be removed.
    """
    for file_path in file_paths:
        if os.path.exists(file_path):
            try:
                os.remove(file_path)
                logger.debug(f"Cleaned file from the filesystem: {file_path}")
            except Exception as e:
                logger.error(f"Failed to remove {file_path}: {str(e)}")
        else:
            logger.debug(
                f"File not found. Unable to clean it from the filesystem: {file_path}"
            )


async def download_pdf(pdf_url: str, file_path: str) -> str:
    """
    Download the PDF file from the given URL and save it to the specified path.

    pdf_url: str - The URL of the PDF file to download.
    file_path: str - The local path to save the downloaded PDF.

    returns: str - The file path of the downloaded PDF if successful, otherwise an error message.
    raises: Exception - If an error occurs during the download process.
    """
    try:
        logger.info(f"Downloading PDF from: {pdf_url} to: {file_path}")
        async with httpx.AsyncClient() as client:
            response = await client.get(pdf_url)
            response.raise_for_status()  # Ensure the request was successful
        with open(file_path, "wb") as pdf_file:
            pdf_file.write(response.content)
        return file_path
    # except httpx.HTTPStatusError as e:
    #     raise e
    except Exception as e:
        raise e



import asyncio
import inspect

from typing_extensions import Annotated

from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.core.skills.click_using_selector import do_click
from sentient.core.skills.enter_text_using_selector import do_entertext
from sentient.core.skills.press_key_combination import do_press_key_combination
from sentient.utils.logger import logger


async def enter_text_and_click(
    text_selector: Annotated[
        str,
        "The properly formatted DOM selector query, for example [mmid='1234'], where the text will be entered. Use mmid attribute. mmid will always be a number",
    ],
    text_to_enter: Annotated[
        str,
        "The text that will be entered into the element specified by text_selector.",
    ],
    click_selector: Annotated[
        str,
        "The properly formatted DOM selector query, for example [mmid='1234'], for the element that will be clicked after text entry. mmid will always be a number",
    ],
    wait_before_click_execution: Annotated[
        float, "Optional wait time in seconds before executing the click.", float
    ],
) -> Annotated[
    str, "A message indicating success or failure of the text entry and click."
]:
    """
    Enters text into an element and then clicks on another element.

    Parameters:
    - text_selector: The selector for the element to enter text into. It should be a properly formatted DOM selector query, for example [mmid='1234'], where the text will be entered. Use the mmid attribute.
    - text_to_enter: The text to enter into the element specified by text_selector.
    - click_selector: The selector for the element to click. It should be a properly formatted DOM selector query, for example [mmid='1234'].
    - wait_before_click_execution: Optional wait time in seconds before executing the click action. Default is 0.0.

    Returns:
    - A message indicating the success or failure of the text entry and click.

    Raises:
    - ValueError: If no active page is found. The OpenURL command opens a new page.

    Example usage:
    ```
    await enter_text_and_click("[mmid='1234']", "Hello, World!", "[mmid='5678']", wait_before_click_execution=1.5)
    ```
    """
    logger.info(
        f"Entering text '{text_to_enter}' into element with selector '{text_selector}' and then clicking element with selector '{click_selector}'."
    )

    # Initialize PlaywrightManager and get the active browser page
    browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
    page = await browser_manager.get_current_page()
    if page is None:  # type: ignore
        logger.error("No active page found")
        raise ValueError("No active page found. OpenURL command opens a new page.")

    await browser_manager.highlight_element(text_selector, True)

    function_name = inspect.currentframe().f_code.co_name  # type: ignore
    await browser_manager.take_screenshots(f"{function_name}_start", page)

    text_entry_result = await do_entertext(
        page, text_selector, text_to_enter, use_keyboard_fill=True
    )

    # await browser_manager.notify_user(text_entry_result["summary_message"])
    if not text_entry_result["summary_message"].startswith("Success"):
        await browser_manager.take_screenshots(f"{function_name}_end", page)
        return f"Failed to enter text '{text_to_enter}' into element with selector '{text_selector}'. Check that the selctor is valid."

    result = text_entry_result

    # if the text_selector is the same as the click_selector, press the Enter key instead of clicking
    if text_selector == click_selector:
        do_press_key_combination_result = await do_press_key_combination(
            browser_manager, page, "Enter"
        )
        if do_press_key_combination_result:
            result["detailed_message"] += (
                f' Instead of click, pressed the Enter key successfully on element: "{click_selector}".'
            )
            # await browser_manager.notify_user(
            #     f'Pressed the Enter key successfully on element: "{click_selector}".',
            #     message_type=MessageType.ACTION,
            # )
        else:
            result["detailed_message"] += (
                f' Clicking the same element after entering text in it, is of no value. Tried pressing the Enter key on element "{click_selector}" instead of click and failed.'
            )
            # await browser_manager.notify_user(
            #     'Failed to press the Enter key on element "{click_selector}".',
            #     message_type=MessageType.ACTION,
            # )
    else:
        await browser_manager.highlight_element(click_selector, True)

        do_click_result = await do_click(
            page, click_selector, wait_before_click_execution
        )
        result["detailed_message"] += f' {do_click_result["detailed_message"]}'
        # await browser_manager.notify_user(do_click_result["summary_message"])

    await asyncio.sleep(
        0.1
    )  # sleep for 100ms to allow the mutation observer to detect changes

    await browser_manager.take_screenshots(f"{function_name}_end", page)

    return result["detailed_message"]



import asyncio
import inspect

from playwright.async_api import Page  # type: ignore
from typing_extensions import Annotated

from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.utils.dom_mutation_observer import (
    subscribe,  # type: ignore
    unsubscribe,  # type: ignore
)
from sentient.utils.logger import logger


async def press_key_combination(
    key_combination: Annotated[str, "The key to press, e.g., Enter, PageDown etc"],
) -> str:
    """
    Presses a key combination on the current active page managed by PlaywrightManager.

    This function simulates the pressing of a key or a combination of keys on the current active web page.
    The `key_combination` should be a string that represents the keys to be pressed, separated by '+' if it's a combination.
    For example, 'Control+C' to copy or 'Alt+F4' to close a window on Windows.

    Parameters:
    - key_combination (Annotated[str, "The key combination to press, e.g., 'Control+C'."]): The key combination to press, represented as a string. For combinations, use '+' as a separator.

    Raises:
    - ValueError: If no active page is found.

    Returns:
    str: status of the operation expressed as a string
    """

    logger.info(f"Executing press_key_combination with key combo: {key_combination}")
    # Create and use the PlaywrightManager
    browser_manager = PlaywrightManager()
    page = await browser_manager.get_current_page()

    if page is None:  # type: ignore
        raise ValueError("No active page found. OpenURL command opens a new page.")

    # Split the key combination if it's a combination of keys
    keys = key_combination.split("+")

    dom_changes_detected = None

    def detect_dom_changes(changes: str):  # type: ignore
        nonlocal dom_changes_detected
        dom_changes_detected = changes  # type: ignore

    subscribe(detect_dom_changes)
    # If it's a combination, hold down the modifier keys
    for key in keys[:-1]:  # All keys except the last one are considered modifier keys
        await page.keyboard.down(key)

    # Press the last key in the combination
    await page.keyboard.press(keys[-1])

    # Release the modifier keys
    for key in keys[:-1]:
        await page.keyboard.up(key)
    await asyncio.sleep(
        0.1
    )  # sleep for 100ms to allow the mutation observer to detect changes
    unsubscribe(detect_dom_changes)

    if dom_changes_detected:
        return f"Key {key_combination} executed successfully.\n As a consequence of this action, new elements have appeared in view:{dom_changes_detected}. This means that the action is not yet executed and needs further interaction. Get all_fields DOM to complete the interaction."

    # await browser_manager.notify_user(
    #     f"Key {key_combination} executed successfully", message_type=MessageType.ACTION
    # )
    return f"Key {key_combination} executed successfully"


async def do_press_key_combination(
    browser_manager: PlaywrightManager, page: Page, key_combination: str
) -> bool:
    """
    Presses a key combination on the provided page.

    This function simulates the pressing of a key or a combination of keys on a web page.
    The `key_combination` should be a string that represents the keys to be pressed, separated by '+' if it's a combination.
    For example, 'Control+C' to copy or 'Alt+F4' to close a window on Windows.

    Parameters:
    - browser_manager (PlaywrightManager): The PlaywrightManager instance.
    - page (Page): The Playwright page instance.
    - key_combination (str): The key combination to press, represented as a string. For combinations, use '+' as a separator.

    Returns:
    bool: True if success and False if failed
    """

    logger.info(f"Executing press_key_combination with key combo: {key_combination}")
    try:
        function_name = inspect.currentframe().f_code.co_name  # type: ignore
        await browser_manager.take_screenshots(f"{function_name}_start", page)
        # Split the key combination if it's a combination of keys
        keys = key_combination.split("+")

        # If it's a combination, hold down the modifier keys
        for key in keys[
            :-1
        ]:  # All keys except the last one are considered modifier keys
            await page.keyboard.down(key)

        # Press the last key in the combination
        await page.keyboard.press(keys[-1])

        # Release the modifier keys
        for key in keys[:-1]:
            await page.keyboard.up(key)

    except Exception as e:
        logger.error(f'Error executing press_key_combination "{key_combination}": {e}')
        return False

    await browser_manager.take_screenshots(f"{function_name}_end", page)

    return True



from typing_extensions import Annotated

from sentient.core.web_driver.playwright import PlaywrightManager


async def geturl() -> (
    Annotated[str, "Returns the full URL of the current active web site/page."]
):
    """
    Returns the full URL of the current page

    Parameters:

    Returns:
    - Full URL the browser's active page.
    """

    try:
        # Create and use the PlaywrightManager
        browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
        page = await browser_manager.get_current_page()

        if not page:
            raise ValueError("No active page found. OpenURL command opens a new page.")

        await page.wait_for_load_state("domcontentloaded")

        # Get the URL of the current page
        try:
            title = await page.title()
            current_url = page.url
            if len(current_url) > 250:
                current_url = current_url[:250] + "..."
            return f"Current Page: {current_url}, Title: {title}"  # type: ignore
        except:  # noqa: E722
            current_url = page.url
            return f"Current Page: {current_url}"

    except Exception as e:
        raise ValueError(
            "No active page found. OpenURL command opens a new page."
        ) from e



import asyncio
from asyncio import Future
from typing import Dict, List


def async_input(prompt: str) -> Future:  # type: ignore
    """
    Display a prompt to the user and wait for input in an asynchronous manner.

    Parameters:
    - prompt: The message to display to the user.

    Returns:
    - A Future object that will be fulfilled with the user's input.
    """
    loop = asyncio.get_event_loop()
    return loop.run_in_executor(None, input, prompt)


async def answer_questions_over_cli(questions: List[str]) -> Dict[str, str]:
    """
    Asks a question over the command line and awaits the user's response.

    Parameters:
    - questions: A list of questions to ask the user, e.g., ["What is your favorite site?", "What do you want to search for?"].

    Returns:
    - A dictionary where each key is a question and each value is the user's response.
    """
    answers: Dict[str, str] = {}
    print("*********************************")
    for question in questions:
        answers[question] = await async_input("Question: " + str(question) + " : ")
    print("*********************************")
    return answers



import os
import time
from typing import Any, Union, Dict

from playwright.async_api import Page
from typing_extensions import Annotated

from sentient.config.config import SOURCE_LOG_FOLDER_PATH
from sentient.core.web_driver.playwright import PlaywrightManager
from sentient.utils.dom_helper import wait_for_non_loading_dom_state
from sentient.utils.get_detailed_accessibility_tree import do_get_accessibility_info
from sentient.utils.logger import logger


async def get_dom_with_content_type(
    content_type: Annotated[
        str,
        "The type of content to extract: 'text_only': Extracts the innerText of the highest element in the document and responds with text, or 'input_fields': Extracts the text input and button elements in the dom.",
    ],
) -> Annotated[
    Union[Dict[str, Any], str, None],
    "The output based on the specified content type.",
]:
    """
    Retrieves and processes the DOM of the active page in a browser instance based on the specified content type.

    Parameters
    ----------
    content_type : str
        The type of content to extract. Possible values are:
        - 'text_only': Extracts the innerText of the highest element in the document and responds with text.
        - 'input_fields': Extracts the text input and button elements in the DOM and responds with a JSON object.
        - 'all_fields': Extracts all the fields in the DOM and responds with a JSON object.

    Returns
    -------
    Dict[str, Any] | str | None
        The processed content based on the specified content type. This could be:
        - A JSON object for 'input_fields' with just inputs.
        - Plain text for 'text_only'.
        - A minified DOM represented as a JSON object for 'all_fields'.

    Raises
    ------
    ValueError
        If an unsupported content_type is provided.
    """

    logger.info(f"Executing Get DOM Command based on content_type: {content_type}")
    start_time = time.time()
    # Create and use the PlaywrightManager
    browser_manager = PlaywrightManager(browser_type="chromium", headless=False)
    page = await browser_manager.get_current_page()
    if page is None:  # type: ignore
        raise ValueError("No active page found. OpenURL command opens a new page.")

    extracted_data = None
    await wait_for_non_loading_dom_state(
        page, 2000
    )  # wait for the DOM to be ready, non loading means external resources do not need to be loaded
    user_success_message = ""
    if content_type == "all_fields":
        user_success_message = "Fetched all the fields in the DOM"
        extracted_data = await do_get_accessibility_info(page, only_input_fields=False)
    elif content_type == "input_fields":
        logger.debug("Fetching DOM for input_fields")
        extracted_data = await do_get_accessibility_info(page, only_input_fields=True)
        if extracted_data is None:
            return "Could not fetch input fields. Please consider trying with content_type all_fields."
        user_success_message = "Fetched only input fields in the DOM"
    elif content_type == "text_only":
        # Extract text from the body or the highest-level element
        logger.debug("Fetching DOM for text_only")
        text_content = await get_filtered_text_content(page)
        with open(
            os.path.join(SOURCE_LOG_FOLDER_PATH, "text_only_dom.txt"),
            "w",
            encoding="utf-8",
        ) as f:
            f.write(text_content)
        extracted_data = text_content
        user_success_message = "Fetched the text content of the DOM"
    else:
        raise ValueError(f"Unsupported content_type: {content_type}")

    elapsed_time = time.time() - start_time
    logger.info(f"Get DOM Command executed in {elapsed_time} seconds")
    # await browser_manager.notify_user(
    #     user_success_message, message_type=MessageType.ACTION
    # )
    return extracted_data  # type: ignore


async def get_filtered_text_content(page: Page) -> str:
    text_content = await page.evaluate("""
        () => {
            // Array of query selectors to filter out
            const selectorsToFilter = ['#agente-overlay'];

            // Store the original visibility values to revert later
            const originalStyles = [];

            // Hide the elements matching the query selectors
            selectorsToFilter.forEach(selector => {
                const elements = document.querySelectorAll(selector);
                elements.forEach(element => {
                    originalStyles.push({ element: element, originalStyle: element.style.visibility });
                    element.style.visibility = 'hidden';
                });
            });

            // Get the text content of the page
            let textContent = document?.body?.innerText || document?.documentElement?.innerText || "";

            // Get all the alt text from images on the page
            let altTexts = Array.from(document.querySelectorAll('img')).map(img => img.alt);
            altTexts="Other Alt Texts in the page: " + altTexts.join(' ');

            // Revert the visibility changes
            originalStyles.forEach(entry => {
                entry.element.style.visibility = entry.originalStyle;
            });
            textContent=textContent+" "+altTexts;
            return textContent;
        }
    """)
    return text_content



import asyncio
import json
from typing import Callable, List  # noqa: UP035

from playwright.async_api import Page

# Create an event loop
loop = asyncio.get_event_loop()

DOM_change_callback: List[Callable[[str], None]] = []


def subscribe(callback: Callable[[str], None]) -> None:
    DOM_change_callback.append(callback)


def unsubscribe(callback: Callable[[str], None]) -> None:
    DOM_change_callback.remove(callback)


async def add_mutation_observer(page: Page):
    """
    Adds a mutation observer to the page to detect changes in the DOM.
    When changes are detected, the observer calls the dom_mutation_change_detected function in the browser context.
    This changes can be detected by subscribing to the dom_mutation_change_detected function by individual skills.

    Current implementation only detects when a new node is added to the DOM.
    However, in many cases, the change could be a change in the style or class of an existing node (e.g. toggle visibility of a hidden node).
    """

    await page.evaluate("""
        console.log('Adding a mutation observer for DOM changes');
        new MutationObserver((mutationsList, observer) => {
            let changes_detected = [];
            for(let mutation of mutationsList) {
                if (mutation.type === 'childList') {
                    let allAddedNodes=mutation.addedNodes;
                    for(let node of allAddedNodes) {
                        if(node.tagName && !['SCRIPT', 'NOSCRIPT', 'STYLE'].includes(node.tagName) && !node.closest('#agentDriveAutoOverlay')) {
                            let visibility=true;
                            let content = node.innerText.trim();
                            if(visibility && node.innerText.trim()){
                                if(content) {
                                    changes_detected.push({tag: node.tagName, content: content});
                                }
                            }
                        }
                    }
                } else if (mutation.type === 'characterData') {
                    let node = mutation.target;
                    if(node.parentNode && !['SCRIPT', 'NOSCRIPT', 'STYLE'].includes(node.parentNode.tagName) && !node.parentNode.closest('#agentDriveAutoOverlay')) {
                        let visibility=true;
                        let content = node.data.trim();
                        if(visibility && content && window.getComputedStyle(node.parentNode).display !== 'none'){
                            if(content && !changes_detected.some(change => change.content.includes(content))) {
                                changes_detected.push({tag: node.parentNode.tagName, content: content});
                            }
                        }
                    }
                }
            }
            if(changes_detected.length > 0) {
                window.dom_mutation_change_detected(JSON.stringify(changes_detected));
            }
        }).observe(document, {subtree: true, childList: true, characterData: true});
        """)


async def handle_navigation_for_mutation_observer(page: Page):
    await add_mutation_observer(page)


async def dom_mutation_change_detected(changes_detected: str):
    """
    Detects changes in the DOM (new nodes added) and emits the event to all subscribed callbacks.
    The changes_detected is a string in JSON formatt containing the tag and content of the new nodes added to the DOM.

    e.g.  The following will be detected when autocomplete recommendations show up when one types Nelson Mandela on google search
    [{'tag': 'SPAN', 'content': 'nelson mandela wikipedia'}, {'tag': 'SPAN', 'content': 'nelson mandela movies'}]
    """
    changes_detected = json.loads(changes_detected.replace("\t", "").replace("\n", ""))
    if len(changes_detected) > 0:
        # Emit the event to all subscribed callbacks
        for callback in DOM_change_callback:
            # If the callback is a coroutine function
            if asyncio.iscoroutinefunction(callback):
                await callback(changes_detected)
            # If the callback is a regular function
            else:
                callback(changes_detected)



import asyncio
import textwrap
import uuid
from typing import Dict, List

from colorama import Fore, init
from dotenv import load_dotenv
from langsmith import traceable

from sentient.core.agent.base import BaseAgent
from sentient.core.models.models import (
    Action,
    ActionType,
    AgentInput,
    AgentOutput,
    Memory,
    State,
    Task,
)
from sentient.core.skills.click_using_selector import click
from sentient.core.skills.enter_text_using_selector import EnterTextEntry, entertext
from sentient.core.skills.get_dom_with_content_type import get_dom_with_content_type
from sentient.core.skills.get_url import geturl
from sentient.core.skills.open_url import openurl
from sentient.core.skills.enter_text_and_click import enter_text_and_click
from sentient.core.web_driver.playwright import PlaywrightManager

init(autoreset=True)


class Orchestrator:
    def __init__(
        self, state_to_agent_map: Dict[State, BaseAgent], eval_mode: bool = False, model:str = None
    ):
        load_dotenv()
        self.state_to_agent_map = state_to_agent_map
        self.playwright_manager = PlaywrightManager()
        self.eval_mode = eval_mode
        self.shutdown_event = asyncio.Event()
        self.session_id = str(uuid.uuid4())
        self.model = model

    async def start(self):
        print("Starting orchestrator")
        await self.playwright_manager.async_initialize(eval_mode=self.eval_mode)
        print("Browser started and ready")

        # if not self.eval_mode:
        #     await self._command_loop()
    
    @classmethod
    async def invoke(cls, command: str):
        orchestrator = cls()
        await orchestrator.start()
        result = await orchestrator.execute_command(command)
        return result

    async def _command_loop(self):
        while not self.shutdown_event.is_set():
            try:
                command = await self._get_user_input()
                if command.strip().lower() == "exit":
                    await self.shutdown()
                else:
                    await self.execute_command(command)
            except asyncio.CancelledError:
                break
            except Exception as e:
                print(f"An error occurred: {e}")

    async def _get_user_input(self):
        return await asyncio.get_event_loop().run_in_executor(
            None, input, "Enter your command (or type 'exit' to quit) "
        )

    @traceable(run_type="chain", name="execute_command")
    async def execute_command(self, command: str):
        try:
            # Create initial memory
            self.memory = Memory(
                objective=command,
                current_state=State.BASE_AGENT,
                plan=[],
                thought="",
                completed_tasks=[],
                current_task=None,
                final_response=None,
            )
            print(f"Executing command {self.memory.objective}")
            while self.memory.current_state != State.COMPLETED:
                await self._handle_state()
            self._print_final_response()

            if self.eval_mode:
                return self.memory.final_response
            else:
                return
        except Exception as e:
            print(f"Error executing the command {self.memory.objective}: {e}")

    def run(self) -> Memory:
        while self.memory.current_state != State.COMPLETED:
            self._handle_state()

        self._print_final_response()
        return self.memory

    async def _handle_state(self):
        current_state = self.memory.current_state

        if current_state not in self.state_to_agent_map:
            raise ValueError(f"Unhandled state! No agent for {current_state}")
        
        if current_state == State.BASE_AGENT:
            await self._handle_agnet()
        else:
            raise ValueError(f"Unhandled state: {current_state}")


    async def _handle_agnet(self):
        agent = self.state_to_agent_map[State.BASE_AGENT]
        self._print_memory_and_agent(agent.name)

        # repesenting state with dom representation
        dom = await get_dom_with_content_type(content_type="all_fields")
        url = await geturl()

        input_data = AgentInput(
            objective=self.memory.objective,
            completed_tasks=self.memory.completed_tasks,
            current_page_url=str(url),
            current_page_dom=str(dom),
        )

        output: AgentOutput = await agent.run(
            input_data, session_id=self.session_id, model=self.model
        )

        await self._update_memory_from_agent(output)

        print(f"{Fore.MAGENTA}Base Agent Q has updated the memory.")


    async def _update_memory_from_agent(self, agentq_output: AgentOutput):
        if agentq_output.is_complete:
            self.memory.current_state = State.COMPLETED
            self.memory.final_response = agentq_output.final_response
        elif agentq_output.next_task:
            self.memory.current_state = State.BASE_AGENT
            if agentq_output.next_task_actions:
                action_results = await self.handle_agent_actions(
                    agentq_output.next_task_actions
                )
                print("Action results:", action_results)
                flattened_results = "; ".join(action_results)
                agentq_output.next_task.result = flattened_results

            self.memory.completed_tasks.append(agentq_output.next_task)
            self.memory.plan = agentq_output.plan
            self.memory.thought = agentq_output.thought
            current_task_id = len(self.memory.completed_tasks) + 1
            self.memory.current_task = Task(
                id=current_task_id,
                description=agentq_output.next_task.description,
                url=None,
                result=None,
            )
        else:
            raise ValueError("Planner did not provide next task or completion status")

    async def handle_agent_actions(self, actions: List[Action]):
        results = []
        for action in actions:
            if action.type == ActionType.GOTO_URL:
                result = await openurl(url=action.website, timeout=action.timeout or 1)
                print("Action - GOTO")
            elif action.type == ActionType.TYPE:
                entry = EnterTextEntry(
                    query_selector=f"[mmid='{action.mmid}']", text=action.content
                )
                result = await entertext(entry)
                print("Action - TYPE")
            elif action.type == ActionType.CLICK:
                result = await click(
                    selector=f"[mmid='{action.mmid}']",
                    wait_before_execution=action.wait_before_execution or 1,
                )
                print("Action - CLICK")
            elif action.type == ActionType.ENTER_TEXT_AND_CLICK:
                result = await enter_text_and_click(
                    text_selector=f"[mmid='{action.text_element_mmid}']",
                    text_to_enter=action.text_to_enter,
                    click_selector=f"[mmid='{action.click_element_mmid}']",
                    wait_before_click_execution=action.wait_before_click_execution
                    or 1.5,
                )
                print("Action - ENTER TEXT AND CLICK")
            else:
                result = f"Unsupported action type: {action.type}"

            results.append(result)

        return results

    async def shutdown(self):
        print("Shutting down orchestrator!")
        self.shutdown_event.set()
        await self.playwright_manager.stop_playwright()

    def _print_memory_and_agent(self, agent_type: str):
        print(f"{Fore.CYAN}{'='*50}")
        print(f"{Fore.YELLOW}Current State: {Fore.GREEN}{self.memory.current_state}")
        print(f"{Fore.YELLOW}Agent: {Fore.GREEN}{agent_type}")
        print(f"{Fore.YELLOW}Current Thought: {Fore.GREEN}{self.memory.thought}")
        if len(self.memory.plan) == 0:
            print(f"{Fore.YELLOW}Plan:{Fore.GREEN} none")
        else:
            print(f"{Fore.YELLOW}Plan:")
            for task in self.memory.plan:
                print(f"{Fore.GREEN} {task.id}. {task.description}")
        if self.memory.current_task:
            print(
                f"{Fore.YELLOW}Current Task: {Fore.GREEN}{self.memory.current_task.description}"
            )
        if len(self.memory.completed_tasks) == 0:
            print(f"{Fore.YELLOW}Completed Tasks:{Fore.GREEN} none")
        else:
            print(f"{Fore.YELLOW}Completed Tasks:")
            for task in self.memory.completed_tasks:
                status = "✓" if task.result else " "
                print(f"{Fore.GREEN}  [{status}] {task.id}. {task.description}")
        print(f"{Fore.CYAN}{'='*50}")

    def _print_task_result(self, task: Task):
        print(f"{Fore.CYAN}{'='*50}")
        print(f"{Fore.YELLOW}Task Completed: {Fore.GREEN}{task.description}")
        print(f"{Fore.YELLOW}Result:")
        wrapped_result = textwrap.wrap(task.result, width=80)
        for line in wrapped_result:
            print(f"{Fore.WHITE}{line}")
        print(f"{Fore.CYAN}{'='*50}")

    def _print_final_response(self):
        print(f"\n{Fore.GREEN}{'='*50}")
        print(f"{Fore.GREEN}Objective Completed!")
        print(f"{Fore.GREEN}{'='*50}")
        print(f"{Fore.YELLOW}Final Response:")
        wrapped_response = textwrap.wrap(self.memory.final_response, width=80)
        for line in wrapped_response:
            print(f"{Fore.WHITE}{line}")
        print(f"{Fore.GREEN}{'='*50}")
